#include <lcms.h>
#include "common.h"
#include "cmsprofile.h"
#include "cmsimage.h"
#include "imageset.h"
#include "image_copy.h"

struct cmsProfile {
  cmsHPROFILE profile;
};

static void setWhite(cmsCIExyY *white, double white_x, double white_y);
static void setPrimaries(cmsCIExyYTRIPLE *primaries,
                         double red_x, double red_y,
                         double green_x, double green_y,
                         double blue_x, double blue_y);
static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double *gamma_of_file, int num);
static cmsHPROFILE create_profile(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, LPGAMMATABLE *gamma, int num);
static CmsProfile *cmsProfile_alloc(void);
static void cmsProfile_free(CmsProfile *profile);

static void copy_alpha_channel(unsigned char *pdst, const unsigned char *psrc, ImageType type_src, ImageType type_dst, int len);
static cmsHTRANSFORM create_transform(CmsProfile *profile_in, ImageType type_in,
 CmsProfile *profile_out, ImageType type_out, CmsIntent cintent);
static DWORD image_dw_type(Image *image);
static DWORD imageType_dw_type(ImageType type);
static int cmsIntent_intent(CmsIntent cintent);
static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
#if defined DEBUG
static void info_print(cmsHPROFILE profile);
#endif /* DEBUG */

CmsProfile *cmsProfile_open_mem(void *ptr, size_t len)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_mem()\n");
  profile = cmsProfile_alloc();
  if (ptr != NULL) {
    profile->profile = cmsOpenProfileFromMem(ptr, len);
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_file(const char *filename)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_file() filename = '%s'\n", filename);
  profile = cmsProfile_alloc();
  if (filename != NULL) {
    profile->profile = cmsOpenProfileFromFile(filename, "r");
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_sRGB(void)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_sRGB()\n");
  profile = cmsProfile_alloc();
  profile->profile = cmsCreate_sRGBProfile();
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

CmsProfile *cmsProfile_open_chromaticities_gamma(double white_x, double white_y,
                                                 double red_x, double red_y,
                                                 double green_x, double green_y,
                                                 double blue_x, double blue_y,
                                                 double *gamma_of_file, int num)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries;
  DEBUG_PRINT("open_chromaticities_gamma()\n");
  setWhite(&white, white_x, white_y);
  setPrimaries(&primaries, red_x, red_y, green_x, green_y, blue_x, blue_y);
  profile = cmsProfile_alloc();
  profile->profile = get_profile_CIExyY(&white, &primaries, gamma_of_file, num);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

CmsProfile *cmsProfile_open_chromaticities(double white_x, double white_y,
                                           double red_x, double red_y,
                                           double green_x, double green_y,
                                           double blue_x, double blue_y,
                                           unsigned short *gamma_tables[],
                                           int num_tables,
                                           int num_samples)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries;
  LPGAMMATABLE tables[3];
  DEBUG_PRINT("open_chromaticities()\n");
  setWhite(&white, white_x, white_y);
  setPrimaries(&primaries, red_x, red_y, green_x, green_y, blue_x, blue_y);
  for (int i = 0; i < num_tables; i++) {
    tables[i] = cmsAllocGamma(num_samples);
    CopyMemory(tables[i]->GammaTable, gamma_tables[i], num_samples*sizeof(WORD));
  }
  profile = cmsProfile_alloc();
  profile->profile = create_profile(&white, &primaries, tables, num_tables);
  for (int i = 0; i < num_tables; i++) {
    cmsFreeGamma(tables[i]);
  }
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

static void setWhite(cmsCIExyY *white, double white_x, double white_y)
{
  white->x = white_x;
  white->y = white_y;
  white->Y = 1.0;
}

static void setPrimaries(cmsCIExyYTRIPLE *primaries,
                         double red_x, double red_y,
                         double green_x, double green_y,
                         double blue_x, double blue_y)
{
  primaries->Red.x = red_x;
  primaries->Red.y = red_y;
  primaries->Green.x = green_x;
  primaries->Green.y = green_y;
  primaries->Blue.x = blue_x;
  primaries->Blue.y = blue_y;
  primaries->Red.Y = primaries->Green.Y = primaries->Blue.Y = 1.0;
}

CmsProfile *cmsProfile_open_temp(int tempK, double *gamma_of_file, int num)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries = {
    { 0.6400, 0.3300, 1.0 },
    { 0.3000, 0.6000, 1.0 },
    { 0.1500, 0.0600, 1.0 }
  };
  LPcmsCIExyYTRIPLE pp;

  cmsWhitePointFromTemp(tempK, &white);
  profile = cmsProfile_alloc();
  profile->profile = get_profile_CIExyY(&white, &primaries, gamma_of_file, num);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

int cmsProfile_close(CmsProfile *profile)
{
  int rc = FALSE;
  if (profile == NULL) {
    return rc;
  }
  if (cmsCloseProfile(profile->profile)) {
    profile->profile = NULL;
    rc = TRUE;
  }
  cmsProfile_free(profile);
  return rc;
}

static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double *gamma_of_file, int num)
{
  cmsHPROFILE profile;
  LPGAMMATABLE gamma[3];
  int num_samples = 256;

  if (num > 3) {
    num = 3;
  }
  for (int j = 0; j < num; j++) {
    gamma[j] = cmsBuildGamma(num_samples, 1.0/gamma_of_file[j]);
    if (gamma[j] == NULL) {
      for (int i = 0; i < j; i++) {
        cmsFreeGamma(gamma[i]);
      }
      return NULL;
    }
  }
  profile = create_profile(white, primaries, gamma, num);
  for (int i = 0; i < num; i++) {
    cmsFreeGamma(gamma[i]);
  }
  return profile;
}

static cmsHPROFILE create_profile(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, LPGAMMATABLE *gamma, int num)
{
  cmsHPROFILE profile;

  if (num >= 3) {
    profile = cmsCreateRGBProfile(white, primaries, gamma);
  } else {
    profile = cmsCreateGrayProfile(white, gamma[0]);
  }
  return profile;
}

static CmsProfile *cmsProfile_alloc(void)
{
  CmsProfile *p;
  p = XMALLOC(CmsProfile, 1);
  p->profile = NULL;
  return p;
}

static void cmsProfile_free(CmsProfile *profile)
{
  XFREE(profile);
}


Image *image_transform_new(Image *image_in, CmsProfile *profile_in, CmsProfile *profile_out, CmsIntent intent)
{
  Image *image_out;
  unsigned char *data;
  ImageType type_in, type_out;
  int bpp; /* bytes per pixel */
  int w, h, xoff, yoff;

  image_get_size(image_in, &w, &h);
  image_get_offset(image_in, &xoff, &yoff);
  type_in = image_get_type(image_in);
  image_out = image_new();
  switch (type_in) {
  case IMAGE_TYPE_GRAYA_8:
    type_out = IMAGE_TYPE_RGBA_8;
    break;
  case IMAGE_TYPE_GRAYA_16:
    type_out = IMAGE_TYPE_RGBA_16;
    break;
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_RGBA_16:
    type_out = type_in;
    break;
  default:
    type_out = IMAGE_TYPE_UNKNOWN;
    break;
  }
  image_set_type(image_out, type_out);
  bpp = imageType_bytes_per_pixel(type_out);
  data = XMALLOC(unsigned char, w*h*bpp);
  copy_alpha_channel(data, image_get_data(image_in), type_in, type_out, w*h);
  image_set_data(image_out, data);
  image_set_size(image_out, w, h);
  image_set_offset(image_out, xoff, yoff);
  if (type_in == IMAGE_TYPE_INDEXED_RGBA_8) {
    image_transform_palette(image_in, profile_in, image_out, profile_out, intent);
  } else {
    image_transform_data(image_in, profile_in, image_out, profile_out, intent);
  }
  return image_out;
}

static void copy_alpha_channel(unsigned char *pdst, const unsigned char *psrc, ImageType type_src, ImageType type_dst, int len)
{
  DataCopy *copy;
  copy = image_alpha_dataCopy_get(type_src, type_dst);
  if (copy != NULL) {
    (*copy)(pdst, psrc, len);
  }
}

void image_transform_data(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out, CmsIntent intent)
{
  cmsHTRANSFORM transform;
  ImageType type_in, type_out;

  type_in = image_get_type(image_in);
  type_out = image_get_type(image_out);
  transform = create_transform(profile_in, type_in, profile_out, type_out, intent);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform data.\n");
    return;
  }
  transform_data(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

void image_transform_palette(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out, CmsIntent intent)
{
  cmsHTRANSFORM transform;
  ImageType type_in, type_out;

  type_in = image_get_type(image_in);
  type_out = image_get_type(image_out);
  transform = create_transform(profile_in, type_in, profile_out, type_out, intent);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform palette.\n");
    return;
  }
  transform_indexed_colors(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

void cms_data_transform(unsigned char *data_in, ImageType type_in, CmsProfile *profile_in, unsigned char *data_out, ImageType type_out, CmsProfile *profile_out, CmsIntent intent, unsigned int len)
{
  cmsHTRANSFORM transform;

  transform = create_transform(profile_in, type_in, profile_out, type_out, intent);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform data.\n");
    return;
  }
  cmsDoTransform(transform, data_in, data_out, len);
  cmsDeleteTransform(transform);
}

static cmsHTRANSFORM create_transform(CmsProfile *profile_in, ImageType type_in, CmsProfile *profile_out, ImageType type_out, CmsIntent cintent)
{
  cmsHPROFILE pin, pout;
  cmsHTRANSFORM transform;
  int intent;
  DWORD dw_flags = 0;
  DWORD dw_type_in, dw_type_out;

  pin = profile_in->profile;
  pout = profile_out->profile;
  intent = cmsIntent_intent(cintent);
  dw_type_in = imageType_dw_type(type_in);
  dw_type_out = imageType_dw_type(type_out);
  transform = cmsCreateTransform(pin, dw_type_in, pout, dw_type_out,
                                 intent, dw_flags);
  return transform;
}

static DWORD image_dw_type(Image *image)
{
  ImageType type;

  type = image_get_type(image);
  return imageType_dw_type(type);
}

static DWORD imageType_dw_type(ImageType type)
{
  DWORD dw_type;

  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
    dw_type = TYPE_GRAY_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_GRAYA_8:
    dw_type = TYPE_GRAYA_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_RGB_8:
    dw_type = TYPE_RGB_8;
    DEBUG_PRINT("dw_type=%s\n", "RGB_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    dw_type = TYPE_RGBA_8;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_8(INDEX)");
    break;
  case IMAGE_TYPE_GRAY_8:
    dw_type = TYPE_GRAY_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_8");
    break;
  case IMAGE_TYPE_GRAY_16:
    dw_type = TYPE_GRAY_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_16_SE");
    break;
  case IMAGE_TYPE_GRAYA_8:
    dw_type = TYPE_GRAYA_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_8");
    break;
  case IMAGE_TYPE_GRAYA_16:
    dw_type = TYPE_GRAYA_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_16_SE");
    break;
  case IMAGE_TYPE_RGB_8:
    dw_type = TYPE_RGB_8;
    DEBUG_PRINT("dw_type=%s\n", "RGB_8");
    break;
  case IMAGE_TYPE_RGBA_8:
    dw_type = TYPE_RGBA_8;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_8");
    break;
  case IMAGE_TYPE_RGB_16:
    dw_type = TYPE_RGB_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "RGB_16_SE");
    break;
  case IMAGE_TYPE_RGBA_16:
    dw_type = TYPE_RGBA_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_16_SE");
    break;
  case IMAGE_TYPE_CMYK_8:
    dw_type = TYPE_CMYK_8;
    DEBUG_PRINT("dw_type=%s\n", "CMYK_8");
    break;
  case IMAGE_TYPE_CMYK_8_REV:
    dw_type = TYPE_CMYK_8_REV;
    DEBUG_PRINT("dw_type=%s\n", "CMYK_8_REV");
    break;
  default:
    DEBUG_PRINT("dw_type=%s\n", "unknown");
    break;
  }
  return dw_type;
}

static int cmsIntent_intent(CmsIntent cintent)
{
  int intent;

  switch (cintent) {
  case CMS_INTENT_PERCEPTUAL:
    intent = INTENT_PERCEPTUAL;
    break;
  case CMS_INTENT_RELATIVE_COLORIMETRIC:
    intent = INTENT_RELATIVE_COLORIMETRIC;
    break;
  case CMS_INTENT_ABSOLUTE_COLORIMETRIC:
    intent = INTENT_ABSOLUTE_COLORIMETRIC;
    break;
  case CMS_INTENT_SATURATION:
    intent = INTENT_SATURATION;
    break;
  default:
    fprintf(stderr, "Intent is unknown: %d\n", cintent);
    break;
  }
  return intent;
}

static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  unsigned char *data_in, *data_out;
  unsigned char *p_in, *p_out;
  int bpp_in, bpp_out;
  int w_in, h_in, w_out, h_out;
  int xoff_in, yoff_in, xoff_out, yoff_out;

  DEBUG_PRINT("transform_data() --- start.\n");
  image_get_size(image_in, &w_in, &h_in);
  image_get_offset(image_in, &xoff_in, &yoff_in);
  bpp_in = imageType_bytes_per_pixel(image_get_type(image_in));
  if ((bpp_in <= 0) || (w_in <= 0) || (h_in <= 0)) {
    DEBUG_PRINT("bpp_in = %d, w_in = %d, h_in = %d\n", bpp_in, w_in, h_in);
    return;
  }
  DEBUG_PRINT("transform_data(): in check OK.\n");
  image_get_size(image_out, &w_out, &h_out);
  image_get_offset(image_out, &xoff_out, &yoff_out);
  bpp_out = imageType_bytes_per_pixel(image_get_type(image_out));
  if ((bpp_out <= 0) || (w_out <= 0) || (h_out <= 0)) {
    DEBUG_PRINT("bpp_out = %d, w_out = %d, h_out = %d\n", bpp_out, w_out, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): out check OK.\n");
  if ((w_in > w_out) || (h_in > h_out)) {
    DEBUG_PRINT("w_in = %d, w_out = %d, h_in = %d, h_out = %d\n", w_in, w_out, h_in, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): check OK.\n");
  data_in = image_get_data(image_in);
  data_out = image_get_data(image_out);
  p_in = data_in;
  p_out = data_out+(w_out*yoff_in+xoff_in)*bpp_out;
  if ((xoff_in == 0) && (xoff_out == 0) && (w_in == w_out)) {
    cmsDoTransform(transform, p_in, p_out, w_in*h_in);
  } else {
    for (int j = 0; j < h_in; j++) {
      cmsDoTransform(transform, p_in, p_out, w_in);
      p_in += w_in*bpp_in;
      p_out += w_out*bpp_out;
    }
  }
  DEBUG_PRINT("transform_data() --- end.\n");
}

static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  ImageType type;
  int num;

  type = image_get_type(image_in);
  num = image_get_num_indexes(image_in);
  image_set_num_indexes(image_out, num);
  for (int i = 0; i < num; i++) {
    unsigned char din[4];
    unsigned char dout[4];

    image_get_index(image_in, i, din);
    switch (type) {
    case IMAGE_TYPE_INDEXED_GRAYA_8:
      copy_alpha_channel(dout, din, IMAGE_TYPE_GRAYA_8, IMAGE_TYPE_RGBA_8, 1);
      break;
    case IMAGE_TYPE_INDEXED_RGBA_8:
      copy_alpha_channel(dout, din, IMAGE_TYPE_RGBA_8, IMAGE_TYPE_RGBA_8, 1);
      break;
    default:
      dout[3] = 0xff;
      break;
    }
    cmsDoTransform(transform, din, dout, 1);
    image_set_index(image_out, i, dout);
  }
}

#if defined DEBUG
static void info_print(cmsHPROFILE profile)
{
  FILE *fp;

  fp = stderr;
  if (profile == NULL) {
    fprintf(fp, "profile is NULL\n");
    return;
  }
  fprintf(fp, "product name: %s\n", cmsTakeProductName(profile));
  fprintf(fp, "product desc: %s\n", cmsTakeProductDesc(profile));
  fprintf(fp, "manufacturer: %s\n", cmsTakeManufacturer(profile));
  fprintf(fp, "model: %s\n", cmsTakeModel(profile));
  fprintf(fp, "copyright: %s\n", cmsTakeCopyright(profile));
  fprintf(fp, "ICC version: %x\n", cmsGetProfileICCversion(profile));
  fprintf(fp, "ICC header flags: %x\n", cmsTakeHeaderFlags(profile));
  fprintf(fp, "ICC header attributes: %x\n", cmsTakeHeaderAttributes(profile));
  fprintf(fp, "intent field of the header: %x\n", cmsTakeRenderingIntent(profile));
  fprintf(fp, "number of tags: %d\n", cmsGetTagCount(profile));
  fprintf(fp, "\n");
}
#endif /* DEBUG */
