/*
 * simplified drawing board for X window for xlib
 */
#include <X11/Xlib.h>
#include <X11/Xresource.h>  /* Xrm */
#include <X11/Xatom.h>      /* XA_STRING */
#include <X11/Xutil.h>      /* XTextProperty */
#include <X11/cursorfont.h> /* XC_clock */
#include <pwd.h>            /* getpwnam(), getpwuid() */
#include <unistd.h>         /* getuid() */
#include <libgen.h>         /* basename() */
#include "common.h"
#include "dboard.h"
#include "output_color.h"

#define PROGNAME "dummy"


struct dBoard {
  Display *display;
  Window window;
  GC gc;
  Pixmap pixmap;
  int width;
  int height;
  OutputColor *colors_rgb8;
  OutputColor *colors_rgb16;
  Cursor cursor[CURSOR_MAX];
  XrmDatabase resource;
  const char *progname;
  Handler *handler;
  Processor *processor;
};

static int dBoard_get_background(DBoard *board, XColor *screen_def, XColor *exact_def);
static void dBoard_get_window_pos(DBoard *board, XWindowAttributes *xwa);
static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa);

static void dBoard_init_cursor(DBoard *board);
static void dBoard_free_cursor(DBoard *board);

static OutputColor *dBoard_alloc_colors(DBoard *board, int bit_depth);
static unsigned long dBoard_get_colorpixel(DBoard *board, unsigned short r, unsigned short g, unsigned short b);
;
static DBoardImage *dBoardImage_alloc(int xoff, int yoff, int width, int height);
static void dBoardImage_free(DBoardImage *bimage);
static int dBoard_pending(DBoard *board);
static int dBoard_nextEvent(DBoard *board, XEvent *event);
static XrmDatabase dBoard_get_xrmdb(Display *d, Window window_id);
static int dBoard_preparse_options(char **name, char **display, int *argc, char **argv);
static void dBoard_parse_options(DBoard *board, int *argc, char **argv);
static int dBoard_resource(DBoard *board, char *name, char *class, XrmValue *value);

static  XrmOptionDescRec options[] = {
  {"--iconic",      ".iconic",              XrmoptionNoArg,         "on"},
  {"--name",        ".name",                XrmoptionSepArg,        NULL},
  {"--title",       ".title",               XrmoptionSepArg,        NULL},
  {"-T",            ".title",               XrmoptionSepArg,        NULL},
  {"--display",     ".display",             XrmoptionSepArg,        NULL},
  {"-d",            ".display",             XrmoptionSepArg,        NULL},
  {"--geometry",    ".geometry",            XrmoptionSepArg,        NULL},
  {"-g",            ".geometry",            XrmoptionSepArg,        NULL},
  {"--foreground",  "*foreground",          XrmoptionSepArg,        NULL},
  {"-fg",           "*foreground",          XrmoptionSepArg,        NULL},
  {"--background",  "*background",          XrmoptionSepArg,        NULL},
  {"-bg",           "*background",          XrmoptionSepArg,        NULL},
  {"--borderwidth", ".borderWidth",         XrmoptionSepArg,        NULL},
  {"--xrm",         NULL,                   XrmoptionResArg,        NULL},
};

/*
 * Open DBoard.
 */
DBoard *dBoard_open(const char *progname, int *argc, char **argv)
{
  DBoard *board;
  Display *d;
  Window w;
  XrmDatabase res;
  char *name, *display;
  int s; /* screen number */
  XrmValue value;
  int xoff, yoff;
  unsigned int width, height, border;
  unsigned int width_max, height_max;
  XSizeHints hints;

  /* initialize X resources */
  XrmInitialize();
  name = (char *)progname;
  display = NULL;
  dBoard_preparse_options(&name, &display, argc, argv);
  d = XOpenDisplay(display);
  if (d == NULL) {
    return NULL;
  }
  if (name == NULL) {
    name = getenv("RESOURCE_NAME");
    if (name == NULL) {
      if (argc > 0) {
        name = basename(argv[0]);
      } else {
        name = PROGNAME;
      }
    }
  }
  progname = name;
  res = dBoard_get_xrmdb(d, 0);
  board = XMALLOC(DBoard, 1);
  board->display = d;
  board->resource = res;
  board->progname = progname;
  dBoard_parse_options(board, argc, argv);

  /* setup parameters */
  border = 2;
  if (dBoard_resource(board, ".borderWidth", NULL, &value)) {
    sscanf((char *)value.addr, "%u", &border);
  }
  s = XDefaultScreen(d);
  width_max = XDisplayWidth(d, s);
  height_max = XDisplayHeight(d, s);
  width = 100;
  height = 200;
  xoff = 30;
  yoff = 40;
  hints.flags = 0;
  hints.win_gravity = NorthWestGravity;
  if (dBoard_resource(board, ".geometry", NULL, &value)) {
    int rc;
    rc = XParseGeometry((char *)value.addr, &xoff, &yoff, &width, &height);
    DEBUG_PRINT("before:xoff=%d, yoff=%d, width=%d, height=%d, border=%d\n", xoff, yoff, width, height, border);
    if ((rc & XValue) || (rc & YValue)) {
      hints.flags = USPosition;
    }
    if ((rc & XValue) && (rc & XNegative)) {
      hints.win_gravity = NorthEastGravity;
      xoff = width_max-(width+2*border)-abs(xoff);
    }
    if ((rc & YValue) && (rc & YNegative)) {
      hints.win_gravity = (hints.win_gravity == NorthWestGravity) ?
        SouthWestGravity : SouthEastGravity;
      yoff = height_max-(height+2*border)-abs(yoff);
    }
    DEBUG_PRINT("after:xoff=%d, yoff=%d, width=%d, height=%d, border=%d\n", xoff, yoff, width, height, border);
  }
  XColor screen_def, exact_def;
  dBoard_get_background(board, &screen_def, &exact_def);
  unsigned long background = screen_def.pixel;
#if !0
  w = XCreateSimpleWindow(d, RootWindow(d, s), xoff, yoff, width, height,
                          border, WhitePixel(d, s), background);
#else
  XSetWindowAttributes attributes;
  unsigned long valuemask;

  memset(&attributes, 0, sizeof(XSetWindowAttributes));
  attributes.background_pixel = background;
  valuemask = 0;
  w = XCreateWindow(d, RootWindow(d, s), xoff, yoff, width, height,
                    border, CopyFromParent, CopyFromParent, CopyFromParent,
                    valuemask, &attributes);
#endif /* 0 */
  board->window = w;

  char *title = progname;
  if (dBoard_resource(board, ".title", NULL, &value)) {
    title = (char *)value.addr;
  }

  XTextProperty w_name;

  w_name.value = (unsigned char *)title;
  w_name.encoding = XA_STRING;
  w_name.format = 8;
  w_name.nitems = strlen(title);
  hints.x = xoff;
  hints.y = yoff;
  hints.width = width;
  hints.height = height;
  hints.max_width = width_max*2;
  hints.max_height = height_max*2;
  hints.flags |= USSize | PMaxSize | PWinGravity;
  XSetWMProperties(d, w, &w_name, NULL, argv, *argc, &hints, NULL, NULL);
#if 0
  XmbSetWMProperties(d, w, title, NULL, argv, *argc, &hints, NULL, NULL);
  Xutf8SetWMProperties(d, w, title, NULL, argv, *argc, &hints, NULL, NULL);
#endif /* 0 */
  GC gc = XCreateGC(d, w, 0, 0); /* create graphics contexts */
  board->gc = gc;
  board->pixmap = None;
  board->colors_rgb8 = NULL;
  board->colors_rgb16 = NULL;
  dBoard_init_cursor(board);
  return board;
}

/*
 * window information.
 */
void dBoard_get_attributes(DBoard *board, int *depth, int *width, int *height, unsigned short *bg_r, unsigned short *bg_g, unsigned short *bg_b, int *bg_setted)
{
  Display *display;
  int screen;
  XColor screen_def, exact_def;

  display = board->display;
  screen = XDefaultScreen(display);
  *depth = XDefaultDepth(display, screen);
  *width = XDisplayWidth(display, screen);
  *height = XDisplayHeight(display, screen);
  *bg_setted = dBoard_get_background(board, &screen_def, &exact_def);
  if (*bg_setted) {
    *bg_r = screen_def.red;
    *bg_g = screen_def.green;
    *bg_b = screen_def.blue;
  }
}

static int dBoard_get_background(DBoard *board, XColor *screen_def, XColor *exact_def)
{
  XrmValue value;
  Display *d;
  int s;

  d = board->display;
  s = XDefaultScreen(d);
  if (dBoard_resource(board, "*background", NULL, &value)) {
    DEBUG_PRINT("background = %s\n", (char *)value.addr);
    if (XAllocNamedColor(d, XDefaultColormap(d, s), (char *)value.addr, screen_def, exact_def)) {
      return TRUE;
    }
  }
  screen_def->pixel = exact_def->pixel = XBlackPixel(d, s);
  return FALSE;
}

/*
 * returns the x, y, w, h coords of window.
 * x, y are relative to root window the border is not included
 * (x, y map to top-left pixel in window)
 */
static void dBoard_get_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w, child;
  int s;

  d = board->display;
  w = board->window;
  s = XDefaultScreen(d);
  /* get the window width and height */
  XGetWindowAttributes(d, w, xwa);
  /* get the window origin */
  XTranslateCoordinates(d, w, RootWindow(d, s), 0, 0, &xwa->x, &xwa->y, &child);
  DEBUG_PRINT("xwa:(%d, %d) %dx%d border=%d\n",
              xwa->x, xwa->y, xwa->width, xwa->height, xwa->border_width);
}

static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w, child;
  int x, y;

  d = board->display;
  w = board->window;
  x = 0;
  y = 0;
  XMoveWindow(d, w, x, y);
}
#if 0
static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w;

  /* sets window x,y,w,h values */
  XWindowChanges xwc;

  /* Adjust from window origin, to border origin */
  xwc.x = xwa->x - xwa->border_width - ch_offx;
  xwc.y = xwa->y - xwa->border_width - ch_offy;
  if (xwa->border_width == 0) {
    xwa->border_width = bwidth;
  }
  xwc.border_width = xwa->border_width;

  /* if we're less than max size in one axis, allow window manager doohickeys
     on the screen */
  if ((xwa->width < display_width) && (xwc.x < p_offx)) {
    xwc.x = p_offx;
  }
  if ((xwa->height < display_height) && (xwc.y < p_offy)) {
    xwc.y = p_offy;
  }
  /* Try to keep bottom right decorations inside */
  if ((xwc.x+width-debkludge_offx) > display_width) {
    xwc.x = display_width-width+debkludge_offx;
    if (xwc.x < 0) {
      xwc.x = 0;
    }
  }
  if ((xwc.y+height-debkludge_offy) > display_height) {
    xwc.y = display_height-height+debkludge_offy;
    if (xwc.y < 0) {
      xwc.y = 0;
    }
  }
  xwc.width = xwa->width;
  xwc.height = xwa->height;

  DEBUG_PRINT("SWP: xwa=%d,%d %dx%d xwc=%d,%d %dx%d off=%d,%d bw=%d klg=%d,%d\n",
              xwa->x, xwa->y, xwa->width, xwa->height,
              xwc.x, xwc.y, xwc.width, xwc.height, p_offx, p_offy,
              xwa->border_width, kludge_offx, kludge_offy);

  xwc.x += kludge_offx;
  xwc.y += kludge_offy;

  /* move/resize the window */
  XConfigureWindow(d, w,
                   CWX | CWY | CWWidth | CWHeight /*| CWBorderWidth*/, &xwc);
}
#endif /* 0 */

/*
 * set window and icon name.
 */
void dBoard_set_name(DBoard *board, const char *name)
{
  XrmValue value;
  char *winname, *iconname;
  char *title;
  Display *d;
  Window w;
  const char *progname;

  progname = board->progname;
  title = progname;
  if (dBoard_resource(board, ".title", NULL, &value)) {
    title = (char *)value.addr;
  }
  if (name[0] == '\0') {
    winname = XMALLOC(char, strlen(title)+1);
    iconname = XMALLOC(char, strlen(title)+1);
    sprintf(winname, "%s", title);
    sprintf(iconname, "%s", title);
  } else {
    winname = XMALLOC(char, strlen(title)+strlen(": ")+strlen(name)+1);
    iconname = XMALLOC(char, strlen(name)+1);
    sprintf(winname, "%s: %s", title, name);
    sprintf(iconname, "%s", name);
  }
  d = board->display;
  w = board->window;
  XStoreName(d, w, winname);
  XSetIconName(d, w, iconname);
  XFREE(winname);
  XFREE(iconname);
}

/*
 * Show window.
 */
void dBoard_show(DBoard *board)
{
  Display *d;
  Window w;
  Atom protocols;

  d = board->display;
  w = board->window;
  protocols = XInternAtom(d, "WM_DELETE_WINDOW", True);
  XSetWMProtocols(d, w, &protocols, 1);
  XMapWindow(d, w);
#if 0
  XWindowAttributes xwa;
  dBoard_get_window_pos(board, &xwa);
  dBoard_set_window_pos(board, &xwa);
  dBoard_get_window_pos(board, &xwa);
#endif /* 0 */
}

/*
 * close DBoard.
 */
int dBoard_close(DBoard *board)
{
  Display *d;
  Window w;
  int rc = TRUE;

  dBoard_free_cursor(board);
  outputColor_free(board->colors_rgb8);
  outputColor_free(board->colors_rgb16);
  d = board->display;
  w = board->window;
  XFreeGC(d, board->gc); /* free graphics contexts */
  XDestroyWindow(d, w);  /* delete window */
  XFlush(d);             /* flush */
  XrmDestroyDatabase(board->resource);
  XCloseDisplay(d);      /* close connection */
  XFREE(board);
  return rc;
}

/*
 * event loop.
 */
int dBoard_main(DBoard *board, void *data)
{
  int pending, done;
  XEvent event;
  Handler *handler;
  Processor *processor;

  handler = board->handler;
  processor = board->processor;
  pending = done = FALSE;
  while (!done) {
    pending = dBoard_pending(board);
    if (pending > 0) {
      dBoard_nextEvent(board, &event);
      (*handler)(board, &event, &done, data);
    } else {
      (*processor)(board, &done, data);
    }
  }
  return done;
}

/*
 * flushes the output buffer.
 */
int dBoard_flush(DBoard *board)
{
  return XFlush(board->display);
}

void dBoard_reset(DBoard *board, int width, int height)
{
  Display *display;
  Window window;
  int screen;
  int depth;
  Pixmap pixmap;

  display = board->display;
  window = board->window;
  screen = XDefaultScreen(display);
  depth  = XDefaultDepth(display, screen);
  pixmap = board->pixmap;
  if (pixmap != None) {
    XFreePixmap(display, pixmap);
  }
  pixmap = XCreatePixmap(display, window, width, height, depth);
  board->pixmap = pixmap;
  board->width  = width;
  board->height = height;
  XResizeWindow(display, window, width, height);
}

void dBoard_copy(DBoard *board)
{
  XCopyArea(board->display, board->pixmap, board->window, board->gc,
            0, 0, board->width, board->height, 0, 0);
}

Atom dBoard_delete_window(DBoard *board)
{
  return XInternAtom(board->display, "WM_DELETE_WINDOW", True);
}

/*
 * select input event.
 */
int dBoard_setHandler(DBoard *board, Handler *handler, long event_mask)
{
  board->handler = handler;
  return XSelectInput(board->display, board->window, event_mask);
}

/*
 * non-event process.
 */
void dBoard_setProcessor(DBoard *board, Processor *processor)
{
  board->processor = processor;
}

/*
 * define cursor.
 */
void dBoard_define_cursor(DBoard *board, enum cursorShape shape)
{
  switch(shape) {
  case CURSOR_WAIT:
    XDefineCursor(board->display, board->window, board->cursor[shape]);
    break;
  default:
    XUndefineCursor(board->display, board->window);
    break;
  }
}

/*
 * undefine cursor.
 */
void dBoard_undefine_cursor(DBoard *board)
{
  XUndefineCursor(board->display, board->window);
}

/*
 * initialize cursors.
 */
static void dBoard_init_cursor(DBoard *board)
{
  for (int i = 0; i < CURSOR_MAX; i++) {
    board->cursor[i] = None;
  }
  board->cursor[CURSOR_WAIT] = XCreateFontCursor(board->display, XC_watch);
}

/*
 * delete cursors.
 */
static void dBoard_free_cursor(DBoard *board)
{
  XUndefineCursor(board->display, board->window);
  for (int i = 0; i < CURSOR_MAX; i++) {
    XFreeCursor(board->display, board->cursor[i]);
  }
}

/*
 * gets color indexes.
 */
OutputColor *dBoard_get_colors(DBoard *board, int bit_depth)
{
  OutputColor *colors = NULL;

  switch (bit_depth) {
  case 8:
    colors = board->colors_rgb8;
    if (colors == NULL) {
      colors = dBoard_alloc_colors(board, 8);
      board->colors_rgb8 = colors;
    }
    break;
  case 16:
    colors = board->colors_rgb16;
    if (colors == NULL) {
      colors = dBoard_alloc_colors(board, 16);
      board->colors_rgb16 = colors;
    }
    break;
  default:
    fprintf(stderr, "Not supported bit depth %d.", bit_depth);
    break;
  }
  return colors;
}

/*
 * initialize RGB colors.
 */
static OutputColor *dBoard_alloc_colors(DBoard *board, int bit_depth)
{
  Display *display;
  int screen;
  Colormap cmap;
  unsigned int num_palette;
  OutputColor *colors;

  display = board->display;
  screen = XDefaultScreen(display);
  cmap = XDefaultColormap(display, screen);
  colors = outputColor_alloc(bit_depth);
  num_palette = 0x1 << bit_depth;
  for (unsigned int j = 0; j < num_palette; j++) {
    XColor *color;
    unsigned short c = 0x0;
    for (unsigned short i = 0; i < 16/bit_depth; i++) {
      c |= (j << bit_depth*i) & 0xffff;
    }
    color = outputColor_get_xcolor(colors, 0, j);
    color->green = color->blue = 0;
    color->red = c;
    XAllocColor(display, cmap, color);
    color = outputColor_get_xcolor(colors, 1, j);
    color->red = color->blue = 0;
    color->green = c;
    XAllocColor(display, cmap, color);
    color = outputColor_get_xcolor(colors, 2, j);
    color->red = color->green = 0;
    color->blue = c;
    XAllocColor(display, cmap, color);
  }
  return colors;
}

unsigned long dBoard_get_pixel_rgb16(DBoard *board, unsigned char *p)
{
  unsigned short c16[3];

  for (int i = 0; i < 3; i++) {
    c16[i]  = (*(p++) << 8) & 0xff00;
    c16[i] |= *(p++) & 0xff;
  }
  return dBoard_get_colorpixel(board, c16[0], c16[1], c16[2]);
}

static unsigned long dBoard_get_colorpixel(DBoard *board, unsigned short r, unsigned short g, unsigned short b)
{
  Display *display;
  int screen;
  Colormap cmap;
  XColor color;

  display = board->display;
  screen = XDefaultScreen(display);
  cmap = XDefaultColormap(display, screen);
  color.red = r;
  color.green = g;
  color.blue = b;
  XAllocColor(display, cmap, &color);
  return color.pixel;
}

void dBoard_fill_rectangle(DBoard *board, int x, int y, unsigned int width, unsigned int height, unsigned long pixel)
{
  Display *display;
  Pixmap pixmap;
  GC gc;

  display = board->display;
  pixmap = board->pixmap;
  gc = board->gc;
  XSetForeground(display, gc, pixel);
  XFillRectangle(display, pixmap, gc, x, y, width, height);
}

DBoardImage *dBoard_get_image(DBoard *board, int xoff, int yoff, int width, int height)
{
  Display *display;
  Pixmap pixmap;
  DBoardImage *bimage;

  display = board->display;
  pixmap = board->pixmap;
  bimage = dBoardImage_alloc(xoff, yoff, width, height);
  if (bimage == NULL) {
    return NULL;
  }
  bimage->ximage = XGetImage(display, pixmap, xoff, yoff, width, height, AllPlanes, ZPixmap);
  return bimage;
}

int dBoard_put_image(DBoard *board, DBoardImage *bimage)
{
  Display *display;
  Pixmap pixmap;
  GC gc;
  int rc;

  display = board->display;
  pixmap = board->pixmap;
  gc = board->gc;
  rc = XPutImage(display, pixmap, gc, bimage->ximage, 0, 0, bimage->xoff, bimage->yoff, bimage->width, bimage->height);
  return TRUE;
}

static DBoardImage *dBoardImage_alloc(int xoff, int yoff, int width, int height)
{
  DBoardImage *bimage;

  bimage = XMALLOC(DBoardImage, 1);
  if (bimage == NULL) {
    return NULL;
  }
  bimage->ximage = NULL;
  bimage->xoff = xoff;
  bimage->yoff = yoff;
  bimage->width = width;
  bimage->height = height;
  return bimage;
}

static void dBoardImage_free(DBoardImage *bimage)
{
  int rc;

  if (bimage == NULL) {
    return;
  }
  rc = XDestroyImage(bimage->ximage);
  XFREE(bimage);
  return rc;
}

/*
 * returns number of events.
 */
static int dBoard_pending(DBoard *board)
{
  return XPending(board->display);
}

/*
 * copies the first event.
 */
static int dBoard_nextEvent(DBoard *board, XEvent *event)
{
  return XNextEvent(board->display, event);
}

static XrmDatabase dBoard_get_xrmdb(Display *d, Window windowid)
{
  XrmDatabase res;

  /* don't use XResourceManagerString, as it is a snapshot of the
     string when the display was opened, and doesn't change */
  Atom resAtom = XInternAtom(d, "RESOURCE_MANAGER", True);
  if (resAtom != None) {
    Atom actType;
    int  rc, actFormat;
    unsigned long nitems, nleft;
    unsigned char *data;
    Window w;

    w = (windowid != 0) ? windowid : RootWindow(d, 0);
    rc = XGetWindowProperty(d, w, resAtom, 0L, 1L, False,
                            XA_STRING, &actType, &actFormat, &nitems, &nleft,
                            (unsigned char **)&data);
    if ((rc == Success) && (actType == XA_STRING) && (actFormat == 8)) {
      if ((nitems > 0) && (data != NULL)) {
        XFree(data);
      }
      w = (windowid != 0) ? windowid : RootWindow(d, 0);
      rc = XGetWindowProperty(d, w, resAtom, 0L, (long)((nleft+4+3)/4), False,
                              XA_STRING, &actType, &actFormat, &nitems, &nleft,
                              (unsigned char **)&data);
      if ((rc == Success) && (actType == XA_STRING) && (actFormat == 8) &&
          (data != NULL)) {
        res = XrmGetStringDatabase((char *)data);
        XFree(data);
        return res;
      }
    }
  }
  char *dpy = XResourceManagerString(d);
  if (dpy != NULL) {
    res = XrmGetStringDatabase(dpy);
  } else {
    char *xdefaults = "/.Xdefaults";
    char *filename;
    char *homedir = getenv("HOME");
    if (homedir == NULL) {
      struct passwd *pw;
      char *user = getenv("USER");
      if (user == NULL) {
        pw = getpwuid(getuid());
      } else {
        pw = getpwnam(user);
      }
      homedir = (pw != NULL) ? pw->pw_dir : "";
    }
    filename = XMALLOC(char, strlen(homedir)+strlen(xdefaults)+1);
    strcpy(filename, homedir);
    strcat(filename, xdefaults);
    res = XrmGetFileDatabase(filename);
    XFREE(filename);
  }
  return res;
}

static int dBoard_preparse_options(char **name, char **display, int *argc, char **argv)
{
  XrmDatabase db = NULL;
  XrmName name_list[3];
  XrmName class_list[3];
  XrmRepresentation type;
  XrmValue value;
  int num_options = 0;
  int argc_tmp;
  char **argv_tmp;

  argc_tmp = *argc;
  argv_tmp = XMALLOC(char *, argc_tmp);
  memcpy(argv_tmp, argv, sizeof(char *)*argc_tmp);
  num_options = sizeof(options)/sizeof(options[0]);
  XrmParseCommand(&db, options, num_options, ".", &argc_tmp, argv_tmp);
  XFREE(argv_tmp);
  name_list[0] = class_list[0] = XrmPermStringToQuark(".");
  name_list[2] = class_list[2] = NULLQUARK;
  if (name != NULL) {
    name_list[1] = XrmPermStringToQuark("name");
    if (XrmQGetResource(db, name_list, name_list, &type, &value)) {
      *name = value.addr;
    }
  }
  if (display != NULL) {
    name_list[1] = XrmPermStringToQuark("display");
    if (XrmQGetResource(db, name_list, name_list, &type, &value)) {
      *display = value.addr;
    }
  }
  XrmDestroyDatabase(db);
  return TRUE;
}

static void dBoard_parse_options(DBoard *board, int *argc, char **argv)
{
  XrmDatabase db;
  const char *progname;
  int number_of_options;

  db = board->resource;
  progname = board->progname;
  number_of_options = sizeof(options)/sizeof(options[0]);
  XrmParseCommand(&db, options, number_of_options, progname, argc, argv);
}

static int dBoard_resource(DBoard *board, char *name, char *class, XrmValue *value)
{
  XrmDatabase db;
  const char *progname;
  char *type;
  char *name_str;
  Bool rc;

  db = board->resource;
  progname = board->progname;
  name_str = XMALLOC(char, strlen(progname)+strlen(name)+1);
  strcpy(name_str, progname);
  strcat(name_str, name);
  rc = XrmGetResource(db, name_str, class, &type, value);
  if (rc) {
    DEBUG_PRINT(" '%s' // value len %d, addr %p '%s'\n",
                name, value->size, value->addr, value->addr);
  }
  XFREE(name_str);
  return rc;
}
