#include "common.h"
#include "image_blend.h"
#include "image_copy.h"
#include "imageset.h"
#include "cmsprofile.h"
#include "cmsimage.h"

#define IMAGE_BPP_MAX 8 /* RGBA_16 */
#define INDEXED_BPP_MAX 4 /* indexed color length */

/* = c1*(1.0-alpha)+c2*alpha */
#define BLEND_ALPHA(c1, c2, alpha, alpha_max) \
  (c1*(alpha_max-alpha)+c2*alpha)/alpha_max

static void image_blend_index_to_rgb8(Image *dst, Image *src, const unsigned char *bgdata);
static void image_blend_rgb(Image *dst, Image *src, const unsigned char *bgdata);
static void blend_c8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
static void blend_c16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b);
static unsigned short get_c16(const unsigned char *c16);
static void set_c16(unsigned char *c16, unsigned short c);
static unsigned short get_c8(const unsigned char *c8);


void image_blend(Image *dst, Image *src,
                 CmsProfile *profile_in, CmsProfile *profile_out,
                 CmsIntent intent, unsigned char *bgdata)
{
  ImageType type_dst, type_src;

  type_src = image_get_type(src);
  switch (type_src) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
  case IMAGE_TYPE_INDEXED_GRAYA_8:
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    type_dst = image_get_type(dst);
    if (type_dst == IMAGE_TYPE_RGB_8) {
      /* +indexed color to RGB8 */
      /* FIXME */
      /* needless blend for color without alpha */
      image_blend_index_to_rgb8(dst, src, bgdata);
      /* FIXME */
    } else {
      /* +indexed color to indexed color */
      image_copy_data(dst, src);
    }
    break;
  case IMAGE_TYPE_GRAY_8:
  case IMAGE_TYPE_GRAY_16:
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_CMYK_8:
  case IMAGE_TYPE_CMYK_8_REV:
    /* +GRAY or +RGB or +CMYK to RGB */
    if ((profile_in != NULL) && (profile_out != NULL)) {
      image_transform_data(src, profile_in, dst, profile_out, intent);
    } else {
      image_copy_data(dst, src);
    }
    break;
  case IMAGE_TYPE_GRAYA_8:
  case IMAGE_TYPE_GRAYA_16:
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_RGBA_16:
    /* +GRAYA or +RGBA to RGB */
    if ((profile_in != NULL) && (profile_out != NULL)) {
      Image *tmpframe;
      tmpframe = image_transform_new(src, profile_in, profile_out, intent);
      image_blend_rgb(dst, tmpframe, bgdata);
      image_free(tmpframe);
    } else {
      image_blend_rgb(dst, src, bgdata);
    }
    break;
  default:
    break;
  }
}

/*
 * +indexed color to RGB8.
 */
static void image_blend_index_to_rgb8(Image *dst, Image *src, const unsigned char *bgdata)
{
  DataCopy *copy;
  ImageType type_dst, type_src;
  unsigned char *pp1;
  unsigned char *pp2;
  int bpp0; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  type_dst = image_get_type(dst);
  if (type_dst != IMAGE_TYPE_RGB_8) {
    fprintf(stderr, "image type is %d, not RGB8! why?\n", type_dst);
    return;
  }
  type_src = image_get_type(src);
  switch (type_src) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
    copy = image_dataCopy_get(IMAGE_TYPE_GRAY_8, IMAGE_TYPE_RGBA_8);
    break;
  case IMAGE_TYPE_INDEXED_GRAYA_8:
    copy = image_dataCopy_get(IMAGE_TYPE_GRAYA_8, IMAGE_TYPE_RGBA_8);
    break;
  case IMAGE_TYPE_INDEXED_RGB_8:
    copy = image_dataCopy_get(IMAGE_TYPE_RGB_8, IMAGE_TYPE_RGBA_8);
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    copy = NULL;
    break;
  default:
    return;
    break;
  }
  bpp0 = imageType_bytes_per_pixel(type_dst);
  pp1 = image_get_data(dst);
  image_get_size(dst, &w0, &h0);
  pp2 = image_get_data(src);
  image_get_size(src, &w, &h);
  image_get_offset(src, &xoff, &yoff);
  y = yoff;
  if (copy == NULL) {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        unsigned char tmp[INDEXED_BPP_MAX];
        image_get_index(src, *pp2, tmp); /* use source palette */
        blend_c8(pp, tmp, bgdata, 3);
        pp += bpp0;
        pp2++;
      }
      y++;
    }
  } else {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        unsigned char tmp1[INDEXED_BPP_MAX];
        unsigned char tmp2[INDEXED_BPP_MAX];
        image_get_index(src, *pp2, tmp1); /* use source palette */
        (*copy)(tmp2, tmp1, 1);
        blend_c8(pp, tmp2, bgdata, 3);
        pp += bpp0;
        pp2++;
      }
      y++;
    }
  }
}

/*
 * +GRAYA or +RGBA to RGB.
 */
static void image_blend_rgb(Image *dst, Image *src, const unsigned char *bgdata)
{
  DataCopy *copy;
  void (*blend_c)(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
  ImageType type_dst, type_src;
  unsigned char *pp1;
  unsigned char *pp2;
  int bpp0, bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  type_dst = image_get_type(dst);
  type_src = image_get_type(src);
  switch (type_src) {
  case IMAGE_TYPE_GRAYA_8:
    blend_c = blend_c8;
    copy = image_dataCopy_get(type_src, IMAGE_TYPE_RGBA_8);
    break;
  case IMAGE_TYPE_GRAYA_16:
    blend_c = blend_c16;
    copy = image_dataCopy_get(type_src, IMAGE_TYPE_RGBA_16);
    break;
  case IMAGE_TYPE_RGBA_8:
    blend_c = blend_c8;
    copy = NULL;
    break;
  case IMAGE_TYPE_RGBA_16:
    blend_c = blend_c16;
    copy = NULL;
    break;
  default:
    return;
    break;
  }
  bpp0 = imageType_bytes_per_pixel(type_dst);
  bpp = imageType_bytes_per_pixel(type_src);
  pp1 = image_get_data(dst);
  image_get_size(dst, &w0, &h0);
  pp2 = image_get_data(src);
  image_get_size(src, &w, &h);
  image_get_offset(src, &xoff, &yoff);
  y = yoff;
  if (copy == NULL) {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        (*blend_c)(pp, pp2, bgdata, 3);
        pp += bpp0;
        pp2 += bpp;
      }
      y++;
    }
  } else {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        unsigned char tmp[IMAGE_BPP_MAX];
        (*copy)(tmp, pp2, 1);
        (*blend_c)(pp, tmp, bgdata, 3);
        pp += bpp0;
        pp2 += bpp;
      }
      y++;
    }
  }
}

static void blend_c8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num)
{
  unsigned char *pp;
  unsigned char a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*num;
  a = *(p2+num);
  if (a == 0xff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x00) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x00) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < num; i++) {
    unsigned int c1, c2, cc;
    unsigned char c;

    c1 = *(p1++);
    c2 = *(p2++);
    cc = BLEND_ALPHA(c1, c2, a, 0xff);
    c = (unsigned char)cc;
    *(pp++) = c;
  }
}

static void blend_c16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num)
{
  unsigned char *pp;
  unsigned short a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*2*num;
  a = get_c16(p2+2*num);
  if (a == 0xffff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x0000) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x0000) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < num; i++) {
    unsigned long c1, c2, cc;
    unsigned short c;

    c1 = get_c16(p1);
    c2 = get_c16(p2);
    cc = BLEND_ALPHA(c1, c2, a, 0xffff);
    c = (unsigned short)cc;
    set_c16(pp, c);
    p1 += 2;
    p2 += 2;
    pp += 2;
  }
}

static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c8(&rgb[0]);
  *g = get_c8(&rgb[1]);
  *b = get_c8(&rgb[2]);
}

static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c16(&rgb[0]);
  *g = get_c16(&rgb[2]);
  *b = get_c16(&rgb[4]);
}

static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b)
{
  set_c16(&rgb[0], r);
  set_c16(&rgb[2], g);
  set_c16(&rgb[4], b);
}

static unsigned short get_c16(const unsigned char *c16)
{
  unsigned short c;

  c  = (c16[0] << 8) & 0xff00;
  c |= c16[1] & 0xff;
  return c;
}

static void set_c16(unsigned char *c16, unsigned short c)
{
  c16[0] = (c >> 8) & 0xff;
  c16[1] = c & 0xff;
}

static unsigned short get_c8(const unsigned char *c8)
{
  unsigned short cc, c;

  cc = c8[0];
  c  = (cc << 8) & 0xff00;
  c |= cc & 0xff;
  return c;
}
