#include "common.h"
#include "image_copy.h"
#include "imageset.h"

#define IMAGE_TYPE_IS_INDEXED(type)          \
  (((type) == IMAGE_TYPE_INDEXED_GRAY_8)  || \
   ((type) == IMAGE_TYPE_INDEXED_GRAYA_8) || \
   ((type) == IMAGE_TYPE_INDEXED_RGB_8)   || \
   ((type) == IMAGE_TYPE_INDEXED_RGBA_8))


static void copy_gray8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_gray8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_gray16_to_rgb16(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_rgb8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_rgb16_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_cmyk8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_cmyk8rev_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_alpha_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_alpha_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_alpha_rgba8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len);
static void copy_alpha_rgba16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len);


/*
 * Note: dst must be larger then src.
 */
void image_copy_data(Image *dst, Image *src)
{
  ImageType type_src, type_dst;
  unsigned char *pp;
  unsigned char *pdst;
  unsigned char *psrc;
  int w0, h0, w, h;
  int xoff, yoff;
  size_t bpp0, bpp; /* bytes per pixel */

  type_dst = image_get_type(dst);
  pdst = image_get_data(dst);
  image_get_size(dst, &w0, &h0);
  bpp = imageType_bytes_per_pixel(type_dst);
  type_src = image_get_type(src);
  psrc = image_get_data(src);
  image_get_size(src, &w, &h);
  image_get_offset(src, &xoff, &yoff);
  bpp0 = imageType_bytes_per_pixel(type_src);
  pp = pdst+(w0*yoff+xoff)*bpp0;
  if ((type_src == type_dst) ||
      (IMAGE_TYPE_IS_INDEXED(type_src) && IMAGE_TYPE_IS_INDEXED(type_dst))) {
    /* same type */
    if (w == w0) {
      memcpy(pp, psrc, w*h*bpp);
    } else {
      unsigned char *pp2;
      pp2 = psrc;
      for (int j = 0; j < h; j++) {
        memcpy(pp, pp2, w*bpp);
        pp += w0*bpp0;
        pp2 += w*bpp;
      }
    }
  } else {
    DataCopy *copy;
    copy = image_dataCopy_get(type_src, type_dst);
    if (copy != NULL) {
      if (w == w0) {
        (*copy)(pp, psrc, w*h);
      } else {
        unsigned char *pp2;
        pp2 = psrc;
        for (int j = 0; j < h; j++) {
          (*copy)(pp, pp2, w);
          pp += w0*bpp0;
          pp2 += w*bpp;
        }
      }
    }
  }
}

void image_copy_palette(Image *dst, Image *src)
{
  ImageType type_src, type_dst;
  int num;

  num = image_get_num_indexes(src);
  image_set_num_indexes(dst, num);
  type_dst = image_get_type(dst);
  type_src = image_get_type(src);
  if (type_src == type_dst) {
    for (int i = 0; i < num; i++) {
      unsigned char data[4];
      image_get_index(src, i, data);
      image_set_index(dst, i, data);
    }
  } else {
    DataCopy *copy_data;
    copy_data = NULL;
    switch (type_dst) {
    case IMAGE_TYPE_INDEXED_RGB_8:
      if (type_src == IMAGE_TYPE_INDEXED_GRAY_8) {
        copy_data = copy_gray8_to_rgb8;
      }
      break;
    case IMAGE_TYPE_INDEXED_RGBA_8:
      if (type_src == IMAGE_TYPE_INDEXED_GRAY_8) {
        copy_data = copy_gray8_to_rgba8;
      } else if (type_src == IMAGE_TYPE_INDEXED_GRAYA_8) {
        copy_data = copy_graya8_to_rgba8;
      } else if (type_src == IMAGE_TYPE_INDEXED_RGB_8) {
        copy_data = copy_rgb8_to_rgba8;
      }
      break;
    case IMAGE_TYPE_RGB_8:
      /* do nothing */
      break;
    }
    if (copy_data != NULL) {
      for (int i = 0; i < num; i++) {
        unsigned char data_src[4], data_dst[4];
        image_get_index(src, i, data_src);
        (*copy_data)(data_dst, data_src, 1);
        image_set_index(dst, i, data_dst);
      }
    }
  }
}

DataCopy *image_dataCopy_get(ImageType type_src, ImageType type_dst)
{
  DataCopy *copy = NULL;

  switch (type_dst) {
  case IMAGE_TYPE_RGB_8:
    if (type_src == IMAGE_TYPE_GRAY_8) {
      copy = copy_gray8_to_rgb8;
    } else if (type_src == IMAGE_TYPE_RGB_16) {
      copy = copy_rgb16_to_rgb8;
    } else if (type_src == IMAGE_TYPE_CMYK_8) {
      copy = copy_cmyk8_to_rgb8;
    } else if (type_src == IMAGE_TYPE_CMYK_8_REV) {
      copy = copy_cmyk8rev_to_rgb8;
    }
    break;
  case IMAGE_TYPE_RGBA_8:
    if (type_src == IMAGE_TYPE_GRAY_8) {
      copy = copy_gray8_to_rgba8;
    } else if (type_src == IMAGE_TYPE_GRAYA_8) {
      copy = copy_graya8_to_rgba8;
    } else if (type_src == IMAGE_TYPE_RGB_8) {
      copy = copy_rgb8_to_rgba8;
    }
    break;
  case IMAGE_TYPE_RGB_16:
    if (type_src == IMAGE_TYPE_GRAY_16) {
      copy = copy_gray16_to_rgb16;
    }
    break;
  case IMAGE_TYPE_RGBA_16:
    if (type_src == IMAGE_TYPE_GRAYA_16) {
      copy = copy_graya16_to_rgba16;
    }
    break;
  }
#if defined DEBUG
  if (copy == NULL) {
    DEBUG_PRINT(" dataCopy(%d,%d) is NULL.\n", type_src, type_dst);
  } else {
    DEBUG_PRINT(" dataCopy(%d,%d)\n", type_src, type_dst);
  }
#endif /* DEBUG */
  return copy;
}

DataCopy *image_alpha_dataCopy_get(ImageType type_src, ImageType type_dst)
{
  DataCopy *copy = NULL;

  switch (type_dst) {
  case IMAGE_TYPE_RGBA_8:
    if (type_src == IMAGE_TYPE_GRAYA_8) {
      copy = copy_alpha_graya8_to_rgba8;
    } else if (type_src == IMAGE_TYPE_RGBA_8) {
      copy = copy_alpha_rgba8_to_rgba8;
    }
    break;
  case IMAGE_TYPE_RGBA_16:
    if (type_src == IMAGE_TYPE_GRAYA_16) {
      copy = copy_alpha_graya16_to_rgba16;
    } else if (type_src == IMAGE_TYPE_RGBA_16) {
      copy = copy_alpha_rgba16_to_rgba16;
    }
    break;
  }
#if defined DEBUG
  if (copy == NULL) {
    DEBUG_PRINT(" dataCopy(%d,%d) is NULL.\n", type_src, type_dst);
  } else {
    DEBUG_PRINT(" dataCopy(%d,%d)\n", type_src, type_dst);
  }
#endif /* DEBUG */
  return copy;
}

static void copy_gray8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
  }
}

static void copy_gray8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
    *(pdst++) = 0xff;
  }
}

static void copy_gray16_to_rgb16(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_rgb8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = 0xff;
  }
}

static void copy_rgb16_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+2);
    *(pdst++) = *(psrc+4);
    psrc += 6;
  }
}

static void copy_cmyk8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    unsigned int r_c, g_m, b_y, a_k;
    r_c = 0xff-*(psrc++);
    g_m = 0xff-*(psrc++);
    b_y = 0xff-*(psrc++);
    a_k = 0xff-*(psrc++);
    *(pdst++) = (r_c*a_k)/0xff;
    *(pdst++) = (g_m*a_k)/0xff;
    *(pdst++) = (b_y*a_k)/0xff;
  }
}

static void copy_cmyk8rev_to_rgb8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    unsigned int r_c, g_m, b_y, a_k;
    r_c = *(psrc++);
    g_m = *(psrc++);
    b_y = *(psrc++);
    a_k = *(psrc++);
    *(pdst++) = (r_c*a_k)/0xff;
    *(pdst++) = (g_m*a_k)/0xff;
    *(pdst++) = (b_y*a_k)/0xff;
  }
}


static void copy_alpha_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    pdst+=3;
    psrc++;
    *(pdst++) = *(psrc++);
  }
}

static void copy_alpha_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    pdst+=6;
    psrc+=2;
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_alpha_rgba8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    pdst+=3;
    psrc+=3;
    *(pdst++) = *(psrc++);
  }
}

static void copy_alpha_rgba16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, size_t len)
{
  for (size_t i = 0; i < len; i++) {
    pdst+=6;
    psrc+=6;
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}
