#include "common.h"
#include "imageset.h"
#include "image_copy.h"
#include "image_blend.h"
#include "image_utils.h"
#include "cmsprofile.h"
#include "cmsimage.h"

#define NUM_FRAMES 3
#define BGDATA_MAX 8 /* RGBA_16 */
#define BGUSER_MAX 6 /* RGB_16 */
#define INDEXED_MAX 4 /* indexed color length */

#define IMAGE_TYPE_IS_INDEXED(type)          \
  (((type) == IMAGE_TYPE_INDEXED_GRAY_8)  || \
   ((type) == IMAGE_TYPE_INDEXED_GRAYA_8) || \
   ((type) == IMAGE_TYPE_INDEXED_RGB_8)   || \
   ((type) == IMAGE_TYPE_INDEXED_RGBA_8))

struct imageWork {
  ImageType type;
  Image *current;
  Image *previous;
  Image *frame[NUM_FRAMES];
  int num_frames;
  int last_index;
  BgPrefer bg_prefer;
  ImageType bgdata_type;
  ImageType bguser_type;
  unsigned char *bgdata;
  unsigned char *bguser;
  int dispose_op;
  int updated_xoff;
  int updated_yoff;
  int updated_width;
  int updated_height;
  CmsProfile *profile_embedded;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  CmsIntent intent;
  int use_infile;
};

static ImageType get_imageWork_type(ImageSet *images);
static int get_image_background(ImageWork *work, Image *image, unsigned char *pbgdata);
static void imageWork_set_updated_size(ImageWork *work, int width, int height, int xoff, int yoff);
static CmsProfile *imageWork_get_profile_in(ImageWork *work);
static void image_fill_pixel(Image *image, unsigned char *pixel);


void imageWork_init(ImageWork *work, ImageSet *images)
{
  ImageType type;
  size_t size;
  int width, height;

  type = get_imageWork_type(images);
  imageSet_get_size(images, &width, &height);
  size = imageType_bytes_per_pixel(type)*width*height;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    size_t size0;
    int w, h;
    image = work->frame[i];
    image_get_size(image, &w, &h);
    size0 = imageType_bytes_per_pixel(work->type)*w*h;
    if (size != size0) {
      unsigned char *data;
      data = image_get_data(image);
      if (data != NULL) {
        XFREE(data);
      }
      data = XMALLOC(unsigned char, size);
      image_set_data(image, data);
    }
    image_set_type(image, type);
    image_set_size(image, width, height);
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
  }
  work->type = type;
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;

  imageSet_get_background(images, work->bgdata, &(work->bgdata_type));
  work->profile_embedded = imageSet_get_profile(images);
#if defined DEBUG
  DEBUG_PRINT(" bg_prefer = %d\n", work->bg_prefer);
  DEBUG_PRINT(" bgdata_type = %d\n", work->bgdata_type);
  for (int i = 0; i < BGDATA_MAX; i++) {
    DEBUG_PRINT(" bgdata[%d] = %x\n", i, work->bgdata[i]);
  }
  DEBUG_PRINT(" bguser_type = %d\n", work->bguser_type);
  for (int i = 0; i < BGUSER_MAX; i++) {
    DEBUG_PRINT(" bguser[%d] = %x\n", i, work->bguser[i]);
  }
#endif /* DEBUG */
}

void imageWork_get_background(ImageWork *work, unsigned char *data, ImageType *ptype)
{
  BgPrefer bg_prefer;
  Image *image;
  unsigned char bgdata[BGDATA_MAX];
  ImageType type = IMAGE_TYPE_UNKNOWN;

  DEBUG_PRINT(" bg_prefer=%d\n", work->bg_prefer);
  DEBUG_PRINT(" bgdata_type=%d\n", work->bgdata_type);
  DEBUG_PRINT(" bguser_type=%d\n", work->bguser_type);
  bg_prefer = work->bg_prefer;
  if (bg_prefer == BG_USER) {
    type = work->bguser_type;
    if (type != IMAGE_TYPE_UNKNOWN) {
      memcpy(data, work->bguser, imageType_bytes_per_pixel(type));
    }
    *ptype = type;
    return;
  }
  /* otherwise, gets background of image */
  image = work->current;
  if (get_image_background(work, image, bgdata)) {
    unsigned char bgtmp[INDEXED_MAX];
    size_t len;
    type = image_get_type(image);
    switch (type) {
    case IMAGE_TYPE_INDEXED_RGB_8:
    case IMAGE_TYPE_INDEXED_RGBA_8:
      type = IMAGE_TYPE_RGB_8;
      image_get_index(image, bgdata[0], bgtmp);
      len = imageType_bytes_per_pixel(type);
      memcpy(data, bgtmp, len);
      break;
    case IMAGE_TYPE_RGB_8:
    case IMAGE_TYPE_RGB_16:
      len = imageType_bytes_per_pixel(type);
      memcpy(data, bgdata, len);
      break;
    default:
      type = IMAGE_TYPE_UNKNOWN;
      break;
    }
  }
  *ptype = type;
}

void imageWork_set_background(ImageWork *work, unsigned char *data, BgPrefer prefer)
{
  work->bg_prefer = prefer;
  work->bguser_type = IMAGE_TYPE_RGB_16;
  memcpy(work->bguser, data, imageType_bytes_per_pixel(work->bguser_type));
}

static ImageType get_imageWork_type(ImageSet *images)
{
  ImageType type, worktype;
  Image *image;
  int num;

  image = imageSet_get_first(images);
  if (image == NULL) {
    return IMAGE_TYPE_UNKNOWN;
  }
  type = image_get_type(image);
  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
  case IMAGE_TYPE_INDEXED_GRAYA_8:
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    worktype = IMAGE_TYPE_INDEXED_RGB_8;
    num = imageSet_get_num_frames(images);
    for (int j = 0; j < num; j++) {
      int n = image_get_num_indexes(image);
      for (int i = 0; i < n; i++) {
        unsigned char data[INDEXED_MAX];
        unsigned char alpha = 0xff;
        type = image_get_type(image);
        image_get_index(image, i, data);
        if (type == IMAGE_TYPE_INDEXED_GRAYA_8) {
          worktype = IMAGE_TYPE_INDEXED_RGBA_8;
          alpha = data[1];
        } else if (type == IMAGE_TYPE_INDEXED_RGBA_8) {
          worktype = IMAGE_TYPE_INDEXED_RGBA_8;
          alpha = data[3];
        }
        if ((alpha > 0x0) && (alpha < 0xff)) {
          worktype = IMAGE_TYPE_RGB_8;
          DEBUG_PRINT("imageWork: type = %d\n", worktype);
          return worktype;
        }
      }
      image = image_next(image);
    }
    break;
  case IMAGE_TYPE_GRAY_8:
  case IMAGE_TYPE_GRAYA_8:
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_CMYK_8:
  case IMAGE_TYPE_CMYK_8_REV:
    worktype = IMAGE_TYPE_RGB_8;
    break;
  case IMAGE_TYPE_GRAY_16:
  case IMAGE_TYPE_GRAYA_16:
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    worktype = IMAGE_TYPE_RGB_16;
    break;
  default:
    break;
  }
  DEBUG_PRINT("imageWork: type = %d\n", worktype);
  return worktype;
}

ImageWork *imageWork_alloc(void)
{
  ImageWork *work;

  work = XMALLOC(ImageWork, 1);
  work->type = IMAGE_TYPE_UNKNOWN;
  work->num_frames = NUM_FRAMES;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    image = image_new();
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
    work->frame[i] = image;
  }
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;
  work->bg_prefer = BG_IMAGE;
  work->bgdata_type = IMAGE_TYPE_UNKNOWN;
  work->bguser_type = IMAGE_TYPE_UNKNOWN;
  work->bgdata = XMALLOC(unsigned char, BGDATA_MAX);
  work->bguser = XMALLOC(unsigned char, BGUSER_MAX);
  memset(work->bgdata, 0x7f, BGDATA_MAX);
  memset(work->bguser, 0x7f, BGUSER_MAX);
  imageWork_set_updated_size(work, 0, 0, 0, 0);
  work->profile_embedded = NULL;
  imageWork_set_profiles(work, NULL, NULL, CMS_INTENT_PERCEPTUAL, FALSE);
  return work;
}

void imageWork_free(ImageWork *work)
{
  for (int i = 0; i < work->num_frames; i++) {
    image_free(work->frame[i]);
  }
  XFREE(work->bgdata);
  XFREE(work->bguser);
  XFREE(work);
}

void imageWork_update(ImageWork *work, Image *newframe)
{
  Image *image;
  Image *last_frame;
  ImageType newtype;
  int dispose_op0;
  int dispose_op;
  int blend_op;
  unsigned char bgdata[BGDATA_MAX];
  int w, h, xoff, yoff;
  int full_image = FALSE;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  CmsIntent intent;

  newtype = image_get_type(newframe);
  DEBUG_PRINT(" image type: work = %d / new = %d\n", work->type, newtype);
  last_frame = NULL;
  image = work->current;
  image_set_type(image, work->type);
  dispose_op = image_get_dispose_op(newframe);
  if (dispose_op == DISPOSE_OP_PREVIOUS) {
    work->last_index = (work->last_index == 2) ? 1 : 2;
    last_frame = work->frame[work->last_index];
    image_copy_palette(last_frame, image);
    image_copy_data(last_frame, image);
  }
  /* dispose operation */
  dispose_op0 = image_get_dispose_op(image);
  DEBUG_PRINT(" dispose_op type: %d\n", dispose_op0);
  switch (dispose_op0) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    if (get_image_background(work, image, bgdata)) {
      image_fill_pixel(image, bgdata);
    }
    full_image = TRUE;
    break;
  case DISPOSE_OP_PREVIOUS:
    image_copy_palette(image, work->previous);
    image_copy_data(image, work->previous);
    full_image = TRUE;
    break;
  default:
    DEBUG_PRINT(" unknown dipose_op type %d. why?\n", dispose_op0);
    break;
  }
  if (full_image) {
    image_get_size(image, &w, &h);
    image_get_offset(image, &xoff, &yoff);
  } else {
    image_get_size(newframe, &w, &h);
    image_get_offset(newframe, &xoff, &yoff);
  }
  imageWork_set_updated_size(work, w, h, xoff, yoff);
  if (work->type == IMAGE_TYPE_INDEXED_RGBA_8) {
    work->dispose_op = dispose_op0;
  } else {
    work->dispose_op = DISPOSE_OP_NONE;
  }
  image_set_dispose_op(image, dispose_op);
  profile_in = imageWork_get_profile_in(work);
  profile_out = work->profile_out;
  intent = work->intent;
  if ((profile_in != NULL) && (profile_out != NULL)) {
    image_transform_palette(newframe, profile_in, image, profile_out, intent);
  } else {
    image_copy_palette(image, newframe);
  }
  work->previous = last_frame;
  /* blend operation */
  blend_op = image_get_blend_op(newframe);
  DEBUG_PRINT(" blend_op type: %d\n", blend_op);
  switch (blend_op) {
  case BLEND_OP_NONE:
    get_image_background(work, image, bgdata);
    image_blend(image, newframe, profile_in, profile_out, intent, bgdata);
    break;
  case BLEND_OP_SOURCE:
    get_image_background(work, image, bgdata);
    image_blend(image, newframe, profile_in, profile_out, intent, bgdata);
    break;
  case BLEND_OP_OVER:
    image_blend(image, newframe, profile_in, profile_out, intent, NULL);
    break;
  default:
    DEBUG_PRINT(" unknown blend_op type %d. why?\n", blend_op);
    break;
  }
  if (work->type != IMAGE_TYPE_INDEXED_RGBA_8) {
    blend_op = BLEND_OP_NONE;
  }
  image_set_blend_op(image, blend_op);
}

/*
 * Gets background color for image type.
 */
static int get_image_background(ImageWork *work, Image *image, unsigned char *pbgdata)
{
  BgPrefer bg_prefer;
  ImageType bgtype;
  ImageType type;
  unsigned char *bgdata;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  CmsIntent intent;
  int rc = FALSE;

  DEBUG_PRINT(" bg_prefer=%d\n", work->bg_prefer);
  DEBUG_PRINT(" bgdata_type=%d\n", work->bgdata_type);
  DEBUG_PRINT(" bguser_type=%d\n", work->bguser_type);
  bg_prefer = work->bg_prefer;
  if (bg_prefer == BG_USER) {
    bgtype = work->bguser_type;
    bgdata = work->bguser;
  } else if (bg_prefer == BG_IMAGE) {
    bgtype = work->bgdata_type;
    bgdata = work->bgdata;
  } else {
    DEBUG_PRINT(" bg_prefer is unknown.\n");
    return rc;
  }
  if (bgtype == IMAGE_TYPE_UNKNOWN) {
    return rc;
  }
  type = image_get_type(image);
  if (bg_prefer == BG_IMAGE) {
    /* check color profile */
    profile_in = imageWork_get_profile_in(work);
    profile_out = work->profile_out;
    intent = work->intent;
    if ((profile_in != NULL) && (profile_out != NULL)) {
      if (!IMAGE_TYPE_IS_INDEXED(bgtype)) {
        cms_data_transform(bgdata, bgtype, profile_in,
                           pbgdata, type, profile_out, intent, 1);
        rc = TRUE;
      } else {
        unsigned char bgtmp[INDEXED_MAX];
        switch (type) {
        case IMAGE_TYPE_INDEXED_RGB_8:
        case IMAGE_TYPE_INDEXED_RGBA_8:
          pbgdata[0] = bgdata[0];
          rc = TRUE;
          break;
        case IMAGE_TYPE_RGB_8:
        case IMAGE_TYPE_RGB_16:
          /* indexed color to RGB */
          image_get_index(image, bgdata[0], bgtmp);
          cms_data_transform(bgtmp, IMAGE_TYPE_RGB_8, profile_in,
                             pbgdata, type, profile_out, intent, 1);
          rc = TRUE;
          break;
        }
      }
      DEBUG_PRINT(" rc=%d bgtype=%d type=%d\n", rc, bgtype, type);
      return rc;
    }
  }
  if ((type == bgtype) ||
      (IMAGE_TYPE_IS_INDEXED(type) && IMAGE_TYPE_IS_INDEXED(bgtype))) {
    /* same type */
    memcpy(pbgdata, bgdata, imageType_bytes_per_pixel(bgtype));
    rc = TRUE;
  } else {
    DataCopy *copy;
    copy = image_dataCopy_get(bgtype, type);
    if (copy != NULL) {
      (*copy)(pbgdata, bgdata, 1);
      rc = TRUE;
    }
  }
  DEBUG_PRINT(" rc=%d\n", rc);
  return rc;
}

int imageWork_get_dispose_op(ImageWork *work)
{
  return work->dispose_op;
}

Image *imageWork_get_image(ImageWork *work)
{
  return work->current;
}

void imageWork_get_updated_size(ImageWork *work, int *width, int *height, int *xoff, int *yoff)
{
  *width = work->updated_width;
  *height = work->updated_height;
  *xoff = work->updated_xoff;
  *yoff = work->updated_yoff;
}

static void imageWork_set_updated_size(ImageWork *work, int width, int height, int xoff, int yoff)
{
  work->updated_width = width;
  work->updated_height = height;
  work->updated_xoff = xoff;
  work->updated_yoff = yoff;
}

void imageWork_set_profiles(ImageWork *work, CmsProfile *profile_in, CmsProfile *profile_out, CmsIntent intent, int use_infile)
{
  work->profile_in = profile_in;
  work->profile_out = profile_out;
  work->intent = intent;
  work->use_infile = use_infile;
}

static CmsProfile *imageWork_get_profile_in(ImageWork *work)
{
  CmsProfile *profile;

  if (work->use_infile) {
    profile = work->profile_in;
  } else {
    profile = work->profile_embedded;
    if (profile == NULL) {
      profile = work->profile_in;
    }
  }
  return profile;
}

static void image_fill_pixel(Image *image, unsigned char *pixel)
{
  ImageType type;
  unsigned char *pp;
  int w, h, bpp;

  type = image_get_type(image);
  image_get_size(image, &w, &h);
  pp = image_get_data(image);
  bpp = imageType_bytes_per_pixel(type);
  if (bpp == 1) {
    memset(pp, *pixel, w*h);
  } else if (bpp > 1) {
    for (int i = 0; i < w*h; i++) {
      for (int c = 0; c < bpp; c++) {
        *(pp++) = pixel[c];
      }
    }
  }
}
