#include "common.h"
#include "imageset.h"
#include "cmsprofile.h"

#define BACKGROUND_LEN 8
#define INDEXED_COLOR_LEN 4

struct backgroundColor {
  unsigned char data[BACKGROUND_LEN];
  ImageType type;
};
typedef struct backgroundColor BackgroundColor;

struct imageColor {
  unsigned char data[INDEXED_COLOR_LEN];
};
typedef struct imageColor ImageColor;

struct image {
  unsigned char *data; /* image data */
  int xoff, yoff;      /* offset for image */
  int width, height;   /* image size */
  ImageType type;      /* type of image */
  int num_indexes;     /* number of indexed colors */
  ImageColor *colors;  /* color map for indexed palette */
  int delay;           /* interframe delay [milli second] */
  int dispose_op;      /* dispose operation */
  int blend_op;        /* blend operation */
  const char *comment; /* comment text */
  struct image *next;  /* pointer to next image */
};
struct imageSet {
  Image *frames;
  int width, height;   /* images dimensions */
  int num_images;      /* number of images */
  int num_loop;        /* number of loop */
  int num_indexes;     /* number of indexed colors */
  ImageColor *colors;  /* color map for indexed palette */
  BackgroundColor *background;
  CmsProfile *profile; /* color profile */
};

static int imageType_bytes_per_pixel_background(ImageType type);
static int imageType_bytes_per_indexed_color(ImageType type);
static void imageColor_get_color(ImageColor *color, unsigned char *data, int len);
static void imageColor_set_color(ImageColor *color, const unsigned char *data, int len);


ImageSet *imageSet_new(void)
{
  ImageSet *images;

  images = XMALLOC(ImageSet, 1);
  if (images != NULL) {
    imageSet_init(images);
  }
  return images;
}

void imageSet_free(ImageSet *images)
{
  Image *frame;
  int n;

  n = images->num_images;
  frame = images->frames;
  for (int i = 0; i < n; i++) {
    Image *next;
    next = image_next(frame);
    image_free(frame);
    frame = next;
    DEBUG_PRINT("imageSet_free: %d\n", i);
  }
  images->frames = NULL;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_indexes = 0;
  XFREE(images->colors);
  XFREE(images->background);
  cmsProfile_close(images->profile);
}

/*
 * add an image frame.
 */
void imageSet_add_image(ImageSet *images, Image *image)
{
  Image *p;

  p = images->frames;
  if (p == NULL) {
    images->frames = image;
    images->num_images++;
    return;
  }
  while (p->next != NULL) {
    p = p->next;
  }
  p->next = image;
  images->num_images++;
}

/*
 * gets number of images.
 */
int imageSet_get_num_frames(ImageSet *images)
{
  return images->num_images;
}

/*
 * gets first image.
 */
Image *imageSet_get_first(ImageSet *images)
{
  return images->frames;
}

/*
 * gets background color.
 */
int imageSet_get_background(ImageSet *images, unsigned char *p, ImageType *ptype)
{
  BackgroundColor *bg;
  ImageType type;
  bg = images->background;
  *ptype = type = bg->type;
  if (type == IMAGE_TYPE_UNKNOWN) {
    return FALSE;
  }
  memcpy(p, bg->data, imageType_bytes_per_pixel(type));
  return TRUE;
}

/*
 * sets background color.
 */
void imageSet_set_background(ImageSet *images, const unsigned char *p, ImageType type)
{
  BackgroundColor *bg;
  bg = images->background;
  bg->type = type;
  if (type == IMAGE_TYPE_UNKNOWN) {
    return;
  }
  memcpy(bg->data, p, imageType_bytes_per_pixel(type));
}

/*
 * gets screen size.
 */
void imageSet_get_size(ImageSet *images, int *width, int *height)
{
  *width = images->width;
  *height = images->height;
}

/*
 * sets screen size.
 */
void imageSet_set_size(ImageSet *images, int width, int height)
{
  images->width = width;
  images->height = height;
}

/*
 * gets number of loop.
 */
int imageSet_get_num_plays(ImageSet *images)
{
  return images->num_loop;
}

/*
 * sets number of loop.
 */
void imageSet_set_num_plays(ImageSet *images, int num)
{
  images->num_loop = num;
}

/*
 * gets number of index colors.
 */
int imageSet_get_num_indexes(ImageSet *images)
{
  return images->num_indexes;
}

/*
 * sets number of index colors.
 */
void imageSet_set_num_indexes(ImageSet *images, int num)
{
  if (num > 0) {
    if (num != images->num_indexes) {
      XFREE(images->colors);
      images->num_indexes = num;
      images->colors = XMALLOC(ImageColor, num);
    }
  }
}

/*
 * gets indexed color.
 */
void imageSet_get_index(ImageSet *images, int index, unsigned char *data)
{
  ImageColor *colors;

  colors = images->colors;
  if ((index >= 0) && (index < images->num_indexes)) {
    /* FIXME */
    int bpp = imageType_bytes_per_indexed_color(IMAGE_TYPE_INDEXED_RGBA_8);
    if (bpp > 0) {
      imageColor_get_color(&colors[index], data, bpp);
    }
  }
}

/*
 * sets indexed color.
 */
void imageSet_set_index(ImageSet *images, int index, const unsigned char *data)
{
  ImageColor *colors;

  colors = images->colors;
  if ((index >= 0) && (index < images->num_indexes)) {
    /* FIXME */
    int bpp = imageType_bytes_per_indexed_color(IMAGE_TYPE_INDEXED_RGBA_8);
    if (bpp > 0) {
      imageColor_set_color(&colors[index], data, bpp);
    }
  }
}

/*
 * gets image color profile.
 */
CmsProfile *imageSet_get_profile(ImageSet *images)
{
  return images->profile;
}

/*
 * sets image color profile.
 */
void imageSet_set_profile(ImageSet *images, CmsProfile *profile)
{
  images->profile = profile;
}

void imageSet_init(ImageSet *images)
{
  BackgroundColor *bg;

  images->frames = NULL;
  images->width = 320;
  images->height = 200;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_indexes = 0;
  images->colors = NULL;
  bg = XMALLOC(BackgroundColor, 1);
  bg->type = IMAGE_TYPE_UNKNOWN;
  memset(bg->data, 0x7f, BACKGROUND_LEN);
  images->background = bg;
  images->profile = NULL;
}

Image *image_new(void)
{
  Image *image;

  image = XMALLOC(Image, 1);
  if (image == NULL) {
    return NULL;
  }
  image->data = NULL;
  image->xoff = 0;
  image->yoff = 0;
  image->width = 0;
  image->height = 0;
  image->type = IMAGE_TYPE_UNKNOWN;
  image->num_indexes = 0;
  image->colors = NULL;
  image->delay = 0;
  image->dispose_op = DISPOSE_OP_NONE;
  image->blend_op = BLEND_OP_NONE;
  image->comment = NULL;
  image->next = NULL;
  return image;
}

void image_free(Image *image)
{
  unsigned char *data;
  ImageColor *colors;
  char *comment;

  data = image->data;
  colors = image->colors;
  comment = (char *)image->comment;
  XFREE(data);
  XFREE(colors);
  XFREE(comment);
  XFREE(image);
}

/*
 * gets image data.
 */
unsigned char *image_get_data(Image *image)
{
  return image->data;
}

/*
 * sets image data.
 */
void image_set_data(Image *image, unsigned char *data)
{
  image->data = data;
}

/*
 * gets image size.
 */
void image_get_size(Image *image, int *width, int *height)
{
  *width = image->width;
  *height = image->height;
}

/*
 * sets image size.
 */
void image_set_size(Image *image, int width, int height)
{
  image->width = width;
  image->height = height;
}

/*
 * gets image offset.
 */
void image_get_offset(Image *image, int *xoff, int *yoff)
{
  *xoff = image->xoff;
  *yoff = image->yoff;
}

/*
 * sets image offset.
 */
void image_set_offset(Image *image, int xoff, int yoff)
{
  image->xoff = xoff;
  image->yoff = yoff;
}

/*
 * gets number of index colors.
 */
int image_get_num_indexes(Image *image)
{
  return image->num_indexes;
}

/*
 * sets number of index colors.
 */
void image_set_num_indexes(Image *image, int num)
{
  if (num > 0) {
    if (num != image->num_indexes) {
      XFREE(image->colors);
      image->num_indexes = num;
      image->colors = XMALLOC(ImageColor, num);
    }
  }
}

/*
 * gets indexed color.
 */
void image_get_index(Image *image, int index, unsigned char *data)
{
  ImageColor *colors;

  colors = image->colors;
  if ((index >= 0) && (index < image->num_indexes)) {
    int bpp = imageType_bytes_per_indexed_color(image->type);
    if (bpp > 0) {
      imageColor_get_color(&colors[index], data, bpp);
    }
  }
}

/*
 * sets indexed color.
 */
void image_set_index(Image *image, int index, const unsigned char *data)
{
  ImageColor *colors;

  colors = image->colors;
  if ((index >= 0) && (index < image->num_indexes)) {
    int bpp = imageType_bytes_per_indexed_color(image->type);
    if (bpp > 0) {
      imageColor_set_color(&colors[index], data, bpp);
    }
  }
}

/*
 * sets comment.
 */
void image_set_comment(Image *image, const char *comment)
{
  image->comment = comment;
}

/*
 * gets delay [m sec].
 */
int image_get_delay(Image *image)
{
  return image->delay;
}

/*
 * sets delay [m sec].
 */
void image_set_delay(Image *image, int delay)
{
  image->delay = delay;
}

/*
 * gets dispose operation.
 */
int image_get_dispose_op(Image *image)
{
  return image->dispose_op;
}

/*
 * sets dispose operation.
 */
void image_set_dispose_op(Image *image, int dispose_op)
{
  image->dispose_op = dispose_op;
}

/*
 * gets blend operation.
 */
int image_get_blend_op(Image *image)
{
  return image->blend_op;
}

/*
 * sets blend operation.
 */
void image_set_blend_op(Image *image, int blend_op)
{
  image->blend_op = blend_op;
}

/*
 * gets image type.
 */
ImageType image_get_type(Image *image)
{
  return image->type;
}

/*
 * sets image type.
 */
void image_set_type(Image *image, ImageType type)
{
  image->type = type;
}

/*
 * next image.
 */
Image *image_next(Image *image)
{
  return image->next;
}

/*
 * size of image type.
 */
int imageType_bytes_per_pixel(ImageType type)
{
  int bpp = -1;

  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
  case IMAGE_TYPE_INDEXED_GRAYA_8:
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    bpp = sizeof(unsigned char);
    break;
  case IMAGE_TYPE_GRAY_8:
    bpp = sizeof(unsigned char);
    break;
  case IMAGE_TYPE_GRAY_16:
    bpp = sizeof(unsigned char)*2;
    break;
  case IMAGE_TYPE_GRAYA_8:
    bpp = sizeof(unsigned char)*2;
    break;
  case IMAGE_TYPE_GRAYA_16:
    bpp = sizeof(unsigned char)*4;
    break;
  case IMAGE_TYPE_RGB_8:
    bpp = sizeof(unsigned char)*3;
    break;
  case IMAGE_TYPE_RGB_16:
    bpp = sizeof(unsigned char)*6;
    break;
  case IMAGE_TYPE_RGBA_8:
    bpp = sizeof(unsigned char)*4;
    break;
  case IMAGE_TYPE_RGBA_16:
    bpp = sizeof(unsigned char)*8;
    break;
  case IMAGE_TYPE_CMYK_8:
  case IMAGE_TYPE_CMYK_8_REV:
    bpp = sizeof(unsigned char)*4;
    break;
  default:
    break;
  }
  return bpp;
}

/*
 * size of indexed color.
 */
static int imageType_bytes_per_indexed_color(ImageType type)
{
  int indexType = IMAGE_TYPE_UNKNOWN;

  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
    indexType = IMAGE_TYPE_GRAY_8;
    break;
  case IMAGE_TYPE_INDEXED_GRAYA_8:
    indexType = IMAGE_TYPE_GRAYA_8;
    break;
  case IMAGE_TYPE_INDEXED_RGB_8:
    indexType = IMAGE_TYPE_RGB_8;
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    indexType = IMAGE_TYPE_RGBA_8;
    break;
  default:
    break;
  }
  return imageType_bytes_per_pixel(indexType);
}

static void imageColor_get_color(ImageColor *color, unsigned char *data, int len)
{
  unsigned char *p;
  if (len > INDEXED_COLOR_LEN) {
    len = INDEXED_COLOR_LEN;
  }
  p = color->data;
  for (int i = 0; i < len; i++) {
    *(data++) = *(p++);
  }
}

static void imageColor_set_color(ImageColor *color, const unsigned char *data, int len)
{
  unsigned char *p;
  if (len > INDEXED_COLOR_LEN) {
    len = INDEXED_COLOR_LEN;
  }
  p = color->data;
  for (int i = 0; i < len; i++) {
    *(p++) = *(data++);
  }
}
