/*
 * GIF decoding.
 *
 * GIF type     ImageType
 * (normal)     INDEXED_RGB_8
 * transparent  INDEXED_RGBA_8
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gif_lib.h>
#include "module.h"

static int mod_read_GIF(Entry *entry, ImageSet *images);
static int get_buffer(GifFileType *gif, GifRowType *buffer);
static Image *get_image(GifFileType *gif, GifRowType *buffer, int transparent, int delay, int disposal, char *comment);
static GifRowType *row_buffer_init(GifWord width, GifWord height);
static void row_buffer_free(GifRowType *buffer, GifWord height);
static void set_frame_colormap(Image *image, ColorMapObject *map, int transparent);
static void set_images_colormap(ImageSet *images, ColorMapObject *map);
static int get_dispose_op(int disposal);
static int magic_GIF(const void *buf, size_t length);


static int mod_read_GIF(Entry *entry, ImageSet *images)
{
  GifFileType *gif;
  GifRowType *buffer;
  GifRecordType recordType;
  int loop, transparent, delay, disposal;
  int block;
  char *comment;
  unsigned char bgindex;

  gif = DGifOpenFileName(entry->filename);
  if (gif == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  imageSet_set_size(images, gif->SWidth, gif->SHeight);
  bgindex = gif->SBackGroundColor;
  imageSet_set_background(images, &bgindex, IMAGE_TYPE_INDEXED_RGBA_8);
  buffer = row_buffer_init(gif->SWidth, gif->SHeight);
  if (buffer == NULL) {
    DGifCloseFile(gif);
    return FALSE;
  }

  /* Scan the content of the GIF file and load the image(s) in: */
  loop = -1;
  transparent = -1;
  delay = -1;
  disposal = -1;
  block = 0;
  comment = NULL;
  do {
    int extCode;
    GifByteType *extension;
    Image *image;

    if (DGifGetRecordType(gif, &recordType) == GIF_ERROR) {
      PrintGifError();
      return FALSE;
    }
    switch (recordType) {
    case IMAGE_DESC_RECORD_TYPE:
      if (DGifGetImageDesc(gif) == GIF_ERROR) {
        PrintGifError();
        return FALSE;
      }
      /* Image Position relative to Screen. */
      if ((gif->Image.Left+gif->Image.Width > gif->SWidth) ||
          (gif->Image.Top+gif->Image.Height > gif->SHeight)) {
        fprintf(stderr, "Image#%d is not confined to screen dimension, aborted.\n", imageSet_get_num_frames(images));
        return FALSE;
      }
      if (!get_buffer(gif, buffer)) {
        PrintGifError();
        return FALSE;
      }
      image = get_image(gif, buffer, transparent, delay, disposal, comment);
      if (image == NULL) {
        return FALSE;
      }
      imageSet_add_image(images, image);
      comment = NULL;
      break;
    case EXTENSION_RECORD_TYPE:
      if (DGifGetExtension(gif, &extCode, &extension) == GIF_ERROR) {
        PrintGifError();
        return FALSE;
      }
      int ext_num = 0;
      int animated = FALSE;
      while (extension != NULL) {
        int blocksize = *extension;
        switch (extCode) {
        case APPLICATION_EXT_FUNC_CODE:
          if (block == 0) {
            if (ext_num == 0) {
              if (blocksize == 0x0b) {
                /* check for Netscape loop extension */
                if (memcmp(extension+1, "NETSCAPE2.0", 11) == 0) {
                  animated = TRUE;
                }
              }
            }
            if (animated && ext_num == 1) {
              if (blocksize == 0x03) {
                loop = (*(extension+2))+(*(extension+3)*256);
              }
            }
          }
          break;
        case GRAPHICS_EXT_FUNC_CODE:
          if (blocksize == 0x04) {
            int packed = *(extension+1);
            delay = (*(extension+2))+(*(extension+3)*256);
            transparent = (packed & 0x01) ? (*(extension+4)) : -1;
            disposal = (packed >> 2) & 0x07;
          }
          break;
        case COMMENT_EXT_FUNC_CODE:
          if (blocksize > 0) {
            comment = (char *)malloc(sizeof(char)*blocksize+1);
            if (comment != NULL) {
              strncpy(comment, extension+1, blocksize);
              comment[blocksize] = '\0';
            }
          }
          break;
        default:
          break;
        }
        if (DGifGetExtensionNext(gif, &extension) == GIF_ERROR) {
          PrintGifError();
          return FALSE;
        }
        ext_num++;
      }
      break;
    case TERMINATE_RECORD_TYPE:
      break;
    default:
      /* Should be traps by DGifGetRecordType. */
      break;
    }
    block++;
  } while (recordType != TERMINATE_RECORD_TYPE);
  row_buffer_free(buffer, gif->SHeight);

  imageSet_set_num_plays(images, loop);
  if (gif->SColorMap != NULL) {
    set_images_colormap(images, gif->SColorMap);
  }
#if defined DEBUG
  fprintf(stderr, "gif: imageSet num_indexes: %d\n", imageSet_get_num_indexes(images));
  fprintf(stderr, "gif: bgIndex: %d\n", bgindex);
  fprintf(stderr, "gif: imageNum: %d, loop: %d\n", imageSet_get_num_frames(images), loop);
#endif /* DEBUG */
  if (DGifCloseFile(gif) == GIF_ERROR) {
    PrintGifError();
  }
  return TRUE;
}

/*
 * gets into buffer.
 */
static int get_buffer(GifFileType *gif, GifRowType *buffer)
{
  const int interlacedOffset[] = { 0, 4, 2, 1 };
  const int interlacedJumps[] = { 8, 8, 4, 2 };
  int row, col, width, height;

  /* Image Position relative to Screen. */
  row = gif->Image.Top;
  col = gif->Image.Left;
  width = gif->Image.Width;
  height = gif->Image.Height;
  if (gif->Image.Interlace) {
    /* Need to perform 4 passes on the images: */
    for (int j = 0; j < 4; j++) {
      for (int i = row+interlacedOffset[j]; i < row+height; i += interlacedJumps[j]) {
        if (DGifGetLine(gif, &buffer[i][col], width) == GIF_ERROR) {
          return FALSE;
        }
      }
    }
  } else {
    for (int i = 0; i < height; i++) {
      if (DGifGetLine(gif, &buffer[row++][col], width) == GIF_ERROR) {
        return FALSE;
      }
    }
  }
  return TRUE;
}

/*
 * gets an image frame.
 */
static Image *get_image(GifFileType *gif, GifRowType *buffer, int transparent, int delay, int disposal, char *comment)
{
  ImageType type;
  Image *image;
  unsigned char *p;
  int row, col, width, height;
  ColorMapObject *colorMap;
  GifColorType *colors;
  unsigned char *data;

  type = (transparent >= 0) ?
    IMAGE_TYPE_INDEXED_RGBA_8 : IMAGE_TYPE_INDEXED_RGB_8;
  width = gif->Image.Width;
  height = gif->Image.Height;
  data = (unsigned char *)malloc(width*height*imageType_bytes_per_pixel(type));
  if (data == NULL) {
    return NULL;
  }
  colorMap = (gif->Image.ColorMap != NULL) ?
    gif->Image.ColorMap : gif->SColorMap;
  colors = colorMap->Colors;
  row = gif->Image.Top;
  col = gif->Image.Left;
  p = data;
  for (int j = row; j < row+height; j++) {
    GifPixelType *line = buffer[j];
    for (int i = col; i < col+width; i++) {
      *p++ = line[i];
    }
  }
  image = image_new();
  if (image == NULL) {
    free(data);
    return NULL;
  }
  image_set_delay(image, delay*10);
  image_set_dispose_op(image, get_dispose_op(disposal));
  if (transparent >= 0) {
    image_set_blend_op(image, BLEND_OP_OVER);
  }
  image_set_data(image, data);
  image_set_offset(image, col, row);
  image_set_size(image, width, height);
  image_set_type(image, type);
  set_frame_colormap(image, colorMap, transparent);
  image_set_comment(image, comment);
#if defined DEBUG
  fprintf(stderr, "gif: %d colors (%d, %d) [%dx%d]\n",
          image_get_num_indexes(image), col, row, width, height);
  fprintf(stderr, "gif: transparent: %d, delay: %d[1/100s], disposal: %d\n",
          transparent, delay, disposal);
#endif /* DEBUG */
  return image;
}

/*
 * initialize and set background color.
 */
static GifRowType *row_buffer_init(GifWord width, GifWord height)
{
  GifRowType *buffer;
  size_t size;

  buffer = (GifRowType *)malloc(height*sizeof(GifRowType *));
  if (buffer == NULL) {
    return NULL;
  }
  /* initialize and set background color */
  size = width*sizeof(GifPixelType);
  buffer[0] = (GifRowType)malloc(size);
  if (buffer[0] == NULL) {
    row_buffer_free(buffer, 0);
    return NULL;
  }
  for (int j = 1; j < height; j++) {
    buffer[j] = (GifRowType)malloc(size);
    if (buffer[j] == NULL) {
      row_buffer_free(buffer, j);
      return NULL;
    }
  }
  return buffer;
}

static void row_buffer_free(GifRowType *buffer, GifWord height)
{
  for (int i = 0; i < height; i++) {
    free(buffer[i]);
  }
  free(buffer);
}

static void set_frame_colormap(Image *image, ColorMapObject *map, int transparent)
{
  GifColorType *colors;
  int count;

  colors = map->Colors;
  count = map->ColorCount;
  image_set_num_indexes(image, count);
  for (int i = 0; i < count; i++) {
    unsigned char data[4];
    data[0] = colors->Red;
    data[1] = colors->Green;
    data[2] = colors->Blue;
    data[3] = (i == transparent) ? 0x0 : 0xff;
    image_set_index(image, i, data);
    colors++;
  }
}

static void set_images_colormap(ImageSet *images, ColorMapObject *map)
{
  GifColorType *colors;
  int count;

  colors = map->Colors;
  count = map->ColorCount;
  imageSet_set_num_indexes(images, count);
  for (int i = 0; i < count; i++) {
    unsigned char data[4];
    data[0] = colors->Red;
    data[1] = colors->Green;
    data[2] = colors->Blue;
    data[3] = 0xff;
    imageSet_set_index(images, i, data);
    colors++;
  }
}

static int get_dispose_op(int disposal)
{
  int dispose_op = DISPOSE_OP_NONE;
  switch (disposal) {
  case 0:
    dispose_op = DISPOSE_OP_NONE;
    break;
  case 1:
    dispose_op = DISPOSE_OP_NONE;
    break;
  case 2:
    dispose_op = DISPOSE_OP_BACKGROUND;
    break;
  case 3:
    dispose_op = DISPOSE_OP_PREVIOUS;
    break;
  }
  return dispose_op;
}

static int magic_GIF(const void *buf, size_t length)
{
  const char *key = "GIF";
  if (length > 3) {
    length = 3;
  }
  if (memcmp(buf, key, length) == 0) {
    return TRUE;
  }
  return FALSE;
}

Module module_GIF = {
  "gif",
  magic_GIF,
  mod_read_GIF
};
