/*
 * TIFF decoding.
 *
 * photometric bps spp  ImageType
 * RGB         8   3    RGB_8
 * RGB         8   4    RGBA_8
 * SEPARATED   8   4    CMYK_8
 * others               RGBA_8
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>
#include "module.h"
#include "cmsprofile.h"

static int mod_read_TIFF(Entry *entry, ImageSet *images);
static Image *tiff_read(TIFF *tif);
static int tiff_read_strip(TIFF *tif, uint32 width, uint32 height, unsigned char *data);
static int tiff_read_tile(TIFF *tif, uint32 width, uint32 height, unsigned char *data);
static int tiff_read_rgba_image(TIFF *tif, uint32 width, uint32 height, unsigned char *data);
static CmsProfile *get_cms_profile(TIFF *tif);
#if defined DEBUG
static void info_print(TIFF *tif);
#endif /* DEBUG */
static int magic_TIFF(const void *buf, size_t length);


static int mod_read_TIFF(Entry *entry, ImageSet *images)
{
  TIFF *tif = NULL;
  int dircount = 0;
  int width, height;

  tif = TIFFOpen(entry->filename, "r");
  if (tif == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  width = height = -1;
  do {
    Image *image;

    image = tiff_read(tif);
    if (image != NULL) {
      int w, h;
      image_get_size(image, &w, &h);
      if (w > width) {
        width = w;
      }
      if (h > height) {
        height = h;
      }
      imageSet_add_image(images, image);
    }
#if defined DEBUG
    fprintf(stderr, "tiff: image[%d]\n", dircount);
#endif /* DEBUG */
    dircount++;
  } while (TIFFReadDirectory(tif));
  TIFFClose(tif);

  if ((width > 0) && (height > 0)) {
    imageSet_set_size(images, width, height);
  }
  return TRUE;
}

static Image *tiff_read(TIFF *tif)
{
  uint32 width, height;
  uint16 photometric;
  uint16 bps, spp, extra;
  uint16 *info;
  Image *image;
  ImageType type = IMAGE_TYPE_UNKNOWN;
  unsigned char *data;

  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width)) {
    TIFFError(TIFFFileName(tif), "Image does not define its width");
    return NULL;
  }
  if (!TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height)) {
    TIFFError(TIFFFileName(tif), "Image does not define its height");
    return NULL;
  }
#if defined DEBUG
  info_print(tif);
#endif /* DEBUG */
  TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);
  TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bps);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
  TIFFGetField(tif, TIFFTAG_EXTRASAMPLES, &extra, &info);
  switch (photometric) {
  case PHOTOMETRIC_MINISWHITE:
  case PHOTOMETRIC_MINISBLACK:
    /* INDEXED_GRAY_8 or GRAY_8 */
    break;
  case PHOTOMETRIC_PALETTE:
    /* INDEXED_RGB_8 */
    break;
  case PHOTOMETRIC_RGB:
    if (bps == 8) {
      if (spp == 4) {
        /* RGBA_8 */
        type = IMAGE_TYPE_RGBA_8;
      } else if (spp == 3) {
        /* RGB_8 */
        type = IMAGE_TYPE_RGB_8;
      }
    }
    break;
  case PHOTOMETRIC_SEPARATED:
    if (bps == 8) {
      if (spp == 4) {
        /* CMYK_8 */
        type = IMAGE_TYPE_CMYK_8;
      }
    }
    break;
  default:
    break;
  }
  if (type != IMAGE_TYPE_UNKNOWN) {
    data = (unsigned char *)malloc(width*height*imageType_bytes_per_pixel(type));
    if (TIFFIsTiled(tif)) {
      tiff_read_tile(tif, width, height, data);
    } else {
      tiff_read_strip(tif, width, height, data);
    }
    /*profile = get_cms_profile(tif);*/
  } else {
    type = IMAGE_TYPE_RGBA_8;
    data = (unsigned char *)malloc(width*height*imageType_bytes_per_pixel(type));
    tiff_read_rgba_image(tif, width, height, data);
  }
  image = image_new();
  image_set_data(image, data);
  image_set_size(image, width, height);
  image_set_type(image, type);
  return image;
}

static int tiff_read_strip(TIFF *tif, uint32 width, uint32 height, unsigned char *data)
{
  tdata_t buf;
  tsize_t strip_size;
  tsize_t scanline_size;
  tstrip_t strip_max;
  uint32 rowsperstrip;
  uint16 config, spp;
  unsigned char *src, *dst;

  strip_size = TIFFStripSize(tif);
  fprintf(stderr, "tif: strip_size=%d\n", strip_size);
  buf = _TIFFmalloc(strip_size);
  if (buf == NULL) {
    TIFFError(TIFFFileName(tif), "No space for raster buffer");
  }
  strip_max = TIFFNumberOfStrips(tif);
  fprintf(stderr, "tif: strip_max=%d\n", strip_max);
  scanline_size = TIFFScanlineSize(tif);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
  TIFFGetFieldDefaulted(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);
  fprintf(stderr, "tif: rowsperstrip = %d\n", rowsperstrip);
  TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &config);
  fprintf(stderr, "tif: config = %d\n", config);
  if (config == PLANARCONFIG_CONTIG) {
    uint32 row = 0;
    dst = data;
    for (tstrip_t strip = 0; strip < strip_max; strip++) {
      tsize_t result;
      uint32 nrow = ((row+rowsperstrip) > height) ? height-row : rowsperstrip;
      result = TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0),
                                    buf, nrow*scanline_size);
      if (result < 0) {
        TIFFError(TIFFFileName(tif), "Read error on input strip number %d\n", strip);
        return FALSE;
      }
      row += nrow;
      /*
      src = (unsigned char *)buf;
      for (uint32 row_dst = 0; row_dst < nrow; row_dst++) {
        for (uint32 col_dst = 0; col_dst < width; col_dst++) {
          for (uint16 s = 0; s < spp; s++) {
            *(dst++) = *(src++);
          }
        }
      }
      */
      size_t nsrc = nrow*width*spp;
      memcpy(dst, buf, nsrc);
      dst += nsrc;
    }
  } else if (config == PLANARCONFIG_SEPARATE) {
    uint32 row = 0;
    uint16 s = 0;
    dst = data+s;
    for (tstrip_t strip = 0; strip < strip_max; strip++) {
      tsize_t result;
      uint32 nrow = ((row+rowsperstrip) > height) ? height-row : rowsperstrip;
      result = TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, s),
                                    buf, nrow*scanline_size);
      if (result < 0) {
        TIFFError(TIFFFileName(tif), "Read error on input strip number %d\n", strip);
        return FALSE;
      }
      src = (unsigned char *)buf;
      for (uint32 row_dst = 0; row_dst < nrow; row_dst++) {
        for (uint32 col_dst = 0; col_dst < width; col_dst++) {
          *dst = *(src++);
          dst += spp;
        }
      }
      row += nrow;
      if (row >= height) {
        row = 0;
        s++;
        dst = data+s;
      }
    }
  }
  _TIFFfree(buf);
  return TRUE;
}

static int tiff_read_tile(TIFF *tif, uint32 width, uint32 height, unsigned char *data)
{
  tdata_t buf;
  tsize_t tile_size, tile_row_size;
  tsize_t scanline_size;
  ttile_t num_tiles;
  uint32 tile_width, tile_height;
  uint16 config, spp;
  unsigned char *src, *dst;

  tile_size = TIFFTileSize(tif);
  buf = _TIFFmalloc(tile_size);
  if (buf == NULL) {
    TIFFError(TIFFFileName(tif), "No space for raster buffer");
    return FALSE;
  }
  tile_row_size = TIFFTileRowSize(tif);
  num_tiles = TIFFNumberOfTiles(tif);
  fprintf(stderr, "tif: number_of_tiles=%d\n", num_tiles);
  fprintf(stderr, "tif: tile_size=%d\n", TIFFTileSize(tif));
  fprintf(stderr, "tif: tile_row_size=%d\n", tile_row_size);
  scanline_size = TIFFScanlineSize(tif);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
  TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tile_width);
  TIFFGetField(tif, TIFFTAG_TILELENGTH, &tile_height);
  fprintf(stderr, "tif: tile_width=%d, tile_height=%d\n", tile_width, tile_height);
  TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &config);
  fprintf(stderr, "tif: config = %d\n", config);
  if (config == PLANARCONFIG_CONTIG) {
    uint32 row = 0;
    uint32 col = 0;
    uint32 row_size = 0;
    for (ttile_t tile = 0; tile < num_tiles; tile++) {
      tsize_t result;
      result = TIFFReadTile(tif, buf, col, row, 0, 0);
      if (result < 0) {
        TIFFError(TIFFFileName(tif), "Read error on input tile\n");
        return FALSE;
      }
      uint32 nrow = ((row+tile_height) > height) ? height-row : tile_height;
      uint32 nsrc = ((col+tile_width) > width) ? scanline_size-row_size : tile_row_size;
      src = (unsigned char *)buf;
      dst = data+(row*width+col)*spp;
      for (uint32 row_dst = 0; row_dst < nrow; row_dst++) {
        memcpy(dst, src, nsrc);
        src += nsrc;
        dst += width*spp;
      }
      col += tile_width;
      row_size += tile_row_size;
      if (col >= width) {
        col = 0;
        row_size = 0;
        row += tile_height;
      }
    }
  } else if (config == PLANARCONFIG_SEPARATE) {
    uint32 row = 0;
    uint32 col = 0;
    uint32 row_size = 0;
    uint16 s = 0;
    for (ttile_t tile = 0; tile < num_tiles; tile++) {
      tsize_t result;
      result = TIFFReadTile(tif, buf, col, row, 0, s);
      if (result < 0) {
        TIFFError(TIFFFileName(tif), "Read error on input tile\n");
        return FALSE;
      }
      uint32 nrow = ((row+tile_height) > height) ? height-row : tile_height;
      uint32 nsrc = ((col+tile_width) > width) ? scanline_size-row_size : tile_row_size;
      src = (unsigned char *)buf;
      for (uint32 row_dst = 0; row_dst < nrow; row_dst++) {
        dst = data+((row+row_dst)*width+col)*spp;
        for (uint32 col_dst = 0; col_dst < nsrc; col_dst++) {
          *dst = *(src++);
          dst += spp;
        }
      }
      col += tile_width;
      row_size += tile_row_size;
      if (col >= width) {
        col = 0;
        row_size = 0;
        row += tile_height;
        if (row >= height) {
          row = 0;
          s++;
        }
      }
    }
  }
  _TIFFfree(buf);
  return TRUE;
}

static int tiff_read_rgba_image(TIFF *tif, uint32 width, uint32 height, unsigned char *data)
{
  uint32 *raster;
  unsigned char *src, *dst;

  raster = (uint32*)_TIFFmalloc(width*height*sizeof(uint32));
  if (raster == NULL) {
    TIFFError(TIFFFileName(tif), "No space for raster buffer");
    return FALSE;
  }
  /* Read the image in one chunk into an RGBA array */
  if (!TIFFReadRGBAImageOriented(tif, width, height, raster, ORIENTATION_TOPLEFT, 0)) {
    TIFFError(TIFFFileName(tif), "TIFFReadRGBAImage not supported this image");
    _TIFFfree(raster);
    return FALSE;
  }
  src = (unsigned char *)raster;
  dst = data;
  for (uint32 row = 0; row < height; row++) {
    for (uint32 col = 0; col < width; col++) {
      *(dst++) = *(src++); /* red */
      *(dst++) = *(src++); /* green */
      *(dst++) = *(src++); /* blue */
      *(dst++) = *(src++); /* alpha */
    }
  }
  _TIFFfree(raster);
  return TRUE;
}

/*
 * This snippet shows how to deal with Embedded profiles in TIFF
 * files using Sam Leffer's library.
 */
static CmsProfile *get_cms_profile(TIFF *tif)
{
  CmsProfile *profile;
  void *buffer;
  uint32 len;
  float *chr;

  if (TIFFGetField(tif, TIFFTAG_ICCPROFILE, &len, &buffer)) {
    profile = cmsProfile_open_mem(buffer, len);
    return profile;
  }
  if (TIFFGetField(tif, TIFFTAG_PRIMARYCHROMATICITIES, &chr)) {
    float *white;
    double red_x, red_y;
    double green_x, green_y;
    double blue_x, blue_y;
    red_x = chr[0];
    red_y = chr[1];
    green_x = chr[2];
    green_y = chr[3];
    blue_x = chr[4];
    blue_y = chr[5];
    if (TIFFGetField(tif, TIFFTAG_WHITEPOINT, &white)) {
      uint16 bitspersample;
      double white_x, white_y;
      white_x = white[0];
      white_y = white[1];
      if (TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample)) {
        unsigned int num_samples = (0x1 << bitspersample);
        if (TRUE) {
          /* RGB */
          unsigned short *gm[3];
          if (TIFFGetField(tif, TIFFTAG_TRANSFERFUNCTION, &gm[0], &gm[1], &gm[2])) {
            profile = cmsProfile_open_chromaticities(white_x, white_y,
                                                     red_x, red_y,
                                                     green_x, green_y,
                                                     blue_x, blue_y,
                                                     gm, 3, num_samples);
            return profile;
          }
        } else {
          /* grayscale */
          unsigned short *gm[1];
          if (TIFFGetField(tif, TIFFTAG_TRANSFERFUNCTION, &gm[0])) {
            profile = cmsProfile_open_chromaticities(white_x, white_y,
                                                     red_x, red_y,
                                                     green_x, green_y,
                                                     blue_x, blue_y,
                                                     gm, 1, num_samples);
            return profile;
          }
        }
      }
    }
  }
  return NULL;
#if 0
  uint16 unit;
  if (TIFFGetField(tif, TIFFTAG_GRAYRESPONSEUNIT, &unit)) {
    uint16 bitspersample;
    GRAYRESPONSEUNIT_10S;
    GRAYRESPONSEUNIT_100S;
    GRAYRESPONSEUNIT_1000S;
    GRAYRESPONSEUNIT_10000S;
    GRAYRESPONSEUNIT_100000S;
    if (TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample)) {
      ;
    }
    uint16 curve;
    if (TIFFGetField(tif, TIFFTAG_GRAYRESPONSECURVE, &curve)) {
      ;
    }

  }
#endif /* 0 */
}

#if defined DEBUG
static void info_print(TIFF *tif)
{
  FILE *fp = stderr;
  uint32 width, height;
  uint16 *sampleinfo;
  uint16 extrasamples;
  uint16 photometric, samplesperpixel, bitspersample;
  CmsProfile *profile;
  const char *str;

  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width)) {
    TIFFError(TIFFFileName(tif), "Image does not define its width");
  }
  if (!TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height)) {
    TIFFError(TIFFFileName(tif), "Image does not define its height");
  }
  if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
    TIFFError(TIFFFileName(tif), "Image does not define its photometric");
  }
  if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel)) {
    TIFFError(TIFFFileName(tif), "Image does not define its sampleperpixel");
  }
  if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample)) {
    TIFFError(TIFFFileName(tif), "Image does not define its bitspersample");
  }
  extrasamples = 0;
  if (!TIFFGetField(tif, TIFFTAG_EXTRASAMPLES, &extrasamples, &sampleinfo)) {
    TIFFError(TIFFFileName(tif), "Image does not define its extrasamples");
  }
  switch (photometric) {
  case PHOTOMETRIC_MINISWHITE:
    str = "MINISWHITE";
    break;
  case PHOTOMETRIC_MINISBLACK:
    str = "MINISBLACK";
    break;
  case PHOTOMETRIC_RGB: 
    str = "RGB";
    break;
  case PHOTOMETRIC_PALETTE:
    str = "PALETTE";
    break;
  case PHOTOMETRIC_MASK:
    str = "MASK";
    break;
  case PHOTOMETRIC_SEPARATED:
    str = "SEPARATED";
    break;
  case PHOTOMETRIC_YCBCR:
    str = "YCBCR";
    break;
  case PHOTOMETRIC_CIELAB:
    str = "CIELAB";
    break;
  case PHOTOMETRIC_ICCLAB:
    str = "ICCLAB";
    break;
  case PHOTOMETRIC_ITULAB:
    str = "ITULAB";
    break;
  case PHOTOMETRIC_LOGL:
    str = "LOGL";
    break;
  case PHOTOMETRIC_LOGLUV:
    str = "LOGLUV";
    break;
  default:
    str = "unknown";
    break;
  }
  fprintf(fp, "tiff: isTiled=%d\n", TIFFIsTiled(tif));
  fprintf(fp, "tiff: scanline_size=%d, strip_size=%d, tile_size=%d\n",
          TIFFScanlineSize(tif), TIFFStripSize(tif), TIFFTileSize(tif));
  fprintf(fp,
          "tiff: (%dx%d) photometric: %s, sample/pixel: %u, bits/sample: %u\n",
          width, height, str, samplesperpixel, bitspersample);
  if (extrasamples > 0) {
    fprintf(fp, "tiff: extrasamples=%u\n", extrasamples);
    switch (sampleinfo[0]) {
    case EXTRASAMPLE_UNSPECIFIED: /* workaround for some images without */
      str = "UNSPECIFIED";
      break;
    case EXTRASAMPLE_ASSOCALPHA: /* data is pre-multiplied */
      str = "ASSOCALPHA";
      break;
    case EXTRASAMPLE_UNASSALPHA: /* data is not pre-multiplied */
      str = "UNASSALPHA";
      break;
    default:
      str = NULL;
      break;
    }
    if (str != NULL) {
      fprintf(fp, "tiff: extrasample: %s\n", str);
    }
  }
}
#endif /* DEBUG */

static int magic_TIFF(const void *buf, size_t length)
{
  const char *keys[] = { "MM\x00\x2a", "II\x2a\x00" };
  if (length > 4) {
    length = 4;
  }
  for (int i = 0; i < 2; i++) {
    if (memcmp(buf, keys[i], length) == 0) {
      return TRUE;
    }
  }
  return FALSE;
}

Module module_TIFF = {
  "tif",
  magic_TIFF,
  mod_read_TIFF
};
