/*
 * xbm module.
 */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "module.h"

#define BUFSIZE 256
#define TABLESIZE 256

static int mod_read_xbm(Entry *entry, ImageSet *images);
static int sscanf_define(const char *buf, int *width, int *height, int *hx, int *hy);
static int sscanf_bits(const char *buf);
static Image *fget_image(FILE *fp, int table[], int width, int height);
static int next_hex(FILE *fp, int table[]);
static void init_table(int table[]);


static int mod_read_xbm(Entry *entry, ImageSet *images)
{
  FILE *fp;
  int rc = FALSE;

  if ((fp = fopen(entry->filename, "r")) == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return rc;
  }
  int table[TABLESIZE];
  init_table(table);

  char buf[BUFSIZE];
  Image *image;
  int width, height;
  int hx, hy;
  image = NULL;
  width = height = hx = hy = 0;
  while (fgets(buf, BUFSIZE, fp) != NULL) {
    if (sscanf_define(buf, &width, &height, &hx, &hy)) {
      continue;
    }
    if (sscanf_bits(buf)) {
      if ((width <= 0) || (height <= 0)) {
        break;
      }
      image = fget_image(fp, table, width, height);
      if (image == NULL) {
        break;
      }
      rc = TRUE;
      break;
    }
  }
  if (rc) {
    imageSet_set_size(images, width, height);
    imageSet_add_image(images, image);
  }
  fclose(fp);
  return rc;
}

static int sscanf_define(const char *buf, int *width, int *height, int *hx, int *hy)
{
  char name_type[BUFSIZE];
  int value;

  if (sscanf(buf, "#define %s %d", name_type, &value) == 2) {
    char *type;
    type = strrchr(name_type, '_');
    if (type == NULL) {
      type = name_type;
    } else {
      type++;
    }
    if (strcmp("width", type) == 0) {
      *width = value;
    }
    if (strcmp("height", type) == 0) {
      *height = value;
    }
    if (strcmp("hot", type) == 0) {
      if ((type-- == name_type) || (type-- == name_type)) {
        return TRUE;
      }
      if (strcmp("x_hot", type) == 0) {
        *hx = value;
      }
      if (strcmp("y_hot", type) == 0) {
        *hy = value;
      }
    }
    return TRUE;
  }
  return FALSE;
}

static int sscanf_bits(const char *buf)
{
  char name_type[BUFSIZE];
  char *type;

  if (sscanf(buf, "static unsigned char %s = {", name_type) == 1) {
    ;
  } else if (sscanf(buf, "static char %s = {", name_type) == 1) {
    ;
  } else {
    return FALSE;
  }
  type = strrchr(name_type, '_');
  if (type == NULL) {
    type = name_type;
  } else {
    type++;
  }
  if (strcmp("bits[]", type) != 0) {
    return FALSE;
  }
  return TRUE;
}

static Image *fget_image(FILE *fp, int table[], int width, int height)
{
  unsigned char WHITE[] = { 0xff, 0xff };
  unsigned char BLACK[] = { 0x00, 0xff };
  Image *image;
  unsigned char *data;
  unsigned char *ptr;
  int bytes_per_line;
  int size;
  int w;

  image = image_new();
  if (image == NULL) {
    return NULL;
  }
  data = (unsigned char *)malloc(width*height);
  if (data == NULL) {
    return NULL;
  }
  bytes_per_line = (width+7)/8;
  size = bytes_per_line*height;
  ptr = data;
  w = width;
  for (int bytes = 0; bytes < size; bytes++) {
    int hex, n;
    hex = next_hex(fp, table);
    if (hex < 0) {
      free(data);
      return NULL;
    }
    if (w%8 > 0) {
      n = w%8;
    } else {
      n = 8;
    }
    for (int i = 0; i < n; i++) {
      *(ptr++) = (hex >> i) & 0x1;
    }
    w -= n;
    if (w <= 0) {
      w = width;
    }
  }
  image_set_type(image, IMAGE_TYPE_INDEXED_GRAY_8);
  image_set_size(image, width, height);
  image_set_num_indexes(image, 2);
  image_set_index(image, 0, WHITE);
  image_set_index(image, 1, BLACK);
  image_set_data(image, data);
  return image;
}

/*
 * read next hex value in the input stream, return -1 if EOF.
 */
static int next_hex(FILE *fp, int table[])
{
  int value = 0;
  int done = FALSE;
  int gotone = FALSE;

  while (!done) {
    int ch;
    ch = getc(fp);
    if (ch == EOF) {
      value = -1;
      done = TRUE;
    } else {
      ch &= 0xff;
      if (isascii(ch) && isxdigit(ch)) {
        value = (value << 4) + table[ch];
        gotone = TRUE;
      } else if ((table[ch] < 0) && gotone) {
        done = TRUE;
      }
    }
  }
  return value;
}

static void init_table(int table[])
{
  unsigned char ch0[] = "0123456789";
  unsigned char ch1[] = "abcdef";
  unsigned char ch2[] = "ABCDEF";
  unsigned char ch3[] = " ,}\n\t";
  int n;

  for (int i = 0; i < TABLESIZE; i++) {
    table[i] = 0;
  }
  n = strlen(ch0);
  for (int i = 0; i < n; i++) {
    table[ch0[i]] = i;
  }
  n = strlen(ch1);
  for (int i = 0; i < n; i++) {
    table[ch1[i]] = i+10;
  }
  n = strlen(ch2);
  for (int i = 0; i < n; i++) {
    table[ch2[i]] = i+10;
  }
  n = strlen(ch3);
  for (int i = 0; i < n; i++) {
    table[ch3[i]] = -1;
  }
}

static int magic_xbm(const void *buf, size_t length)
{
  const char *key = "#define";
  if (length > 7) {
    length = 7;
  }
  if (memcmp(buf, key, length) == 0) {
    return TRUE;
  }
  return FALSE;
}

Module module_xbm = {
  "xbm",
  magic_xbm,
  mod_read_xbm
};
