#include "common.h"
#include "option.h"

static int is_end(OptionEntry *opt, const char *argv, int *pindex);
static int compare_get_sub(const char *argv, const char *name, char **argsub);
static char *get_param(int argc, char * const argv[], int *pindex, char *argsub);
static int optionEntry_save_param(OptionEntry *opt, const char *param);
static int optionEntry_save_int(OptionEntry *opt, int v);
static int optionEntry_save_arg_string(OptionEntry *opt, const char *arg);
static int optionEntry_save_arg_int(OptionEntry *opt, const char *arg);
static int optionEntry_save_arg_long(OptionEntry *opt, const char *arg);
static int optionEntry_save_arg_float(OptionEntry *opt, const char *arg);
static int optionEntry_save_arg_double(OptionEntry *opt, const char *arg);
static int optionEntry_can_save(OptionEntry *opt);

/*
 * gets options from arguments.
 *
 * in,out : argc         arguments
 *          argv[]
 * in     : opt  options
 */
int option_get(int *argc, char ***argv, OptionEntry *opt)
{
  int rc = FALSE;
  int argc0;

  argc0 = *argc;
  while (argc0 > 0) {
    if (strcmp(*(*argv), "--") == 0) {
      (*argv)++;
      (*argc)--;
      rc = TRUE;
      break;
    }
    rc = option_get_one(argc, argv, opt);
    if (!rc) {
      break;
    }
    if (*argc == argc0) {
      break;
    }
    argc0 = *argc;
  }
  return rc;
}

/*
 * gets options from arguments.
 *
 * in,out : argc         arguments
 *          argv[]
 * in     : opt  options
 */
int option_get_one(int *argc, char ***argv, OptionEntry *opt)
{
  OptionEntry *popt;
  int enable_argsub = TRUE;
  int index = 0;
  char *pargv0;

  if (*argc < 1) {
    return FALSE;
  }
  popt = opt;
  pargv0 = *(*argv);
  while (!is_end(popt, pargv0, &index)) {
    const char *name;
    char *argsub = NULL;
    char *pargv;
    int cmp;

    pargv = pargv0;
    pargv++;
    if (*pargv == '-') {
      pargv++;
      name = popt->long_name;
    } else {
      name = popt->short_name;
    }
    if (name == NULL) {
      popt++;
      continue;
    }
    switch (popt->arg) {
    case OPTION_ARG_NONE:
      cmp = compare_get_sub(pargv, name, NULL);
      break;
    case OPTION_ARG_STRING:
    case OPTION_ARG_INT:
    case OPTION_ARG_LONG:
    case OPTION_ARG_FLOAT:
    case OPTION_ARG_DOUBLE:
      if (enable_argsub) {
        cmp = compare_get_sub(pargv, name, &argsub);
      } else {
        cmp = compare_get_sub(pargv, name, NULL);
      }
      break;
    default:
      /* error bad type */
      return FALSE;
      break;
    }
    if (cmp == 0) {
      int rc;
      char *param = NULL;
      switch (popt->arg) {
      case OPTION_ARG_NONE:
        break;
      case OPTION_ARG_STRING:
      case OPTION_ARG_INT:
      case OPTION_ARG_LONG:
      case OPTION_ARG_FLOAT:
      case OPTION_ARG_DOUBLE:
        param = get_param(*argc, *argv, &index, argsub);
        if (param == NULL) {
          /* error no argument */
          return FALSE;
        }
        break;
      default:
        /* error bad type */
        return FALSE;
        break;
      }
      rc = optionEntry_save_param(popt, param);
      if (!rc) {
        /* error */;
        return FALSE;
      }
      index++;
    }
    popt++;
  }
  (*argv) += index;
  (*argc) -= index;
  return TRUE;
}

static int is_end(OptionEntry *opt, const char *argv, int *pindex)
{
  if (opt == NULL) {
    return TRUE;
  }
  if ((opt->long_name == NULL) && (opt->short_name == NULL)) {
    return TRUE;
  }
  if (*argv != '-') {
    return TRUE;
  }
  if (strcmp(argv, "--") == 0) {
    (*pindex)++;
    return TRUE;
  }
  return FALSE;
}

static int compare_get_sub(const char *argv, const char *name, char **argsub)
{
  size_t name_len;
  size_t argv_len;
  int cmp;

  if (argsub == NULL) {
    return strcmp(argv, name);
  }
  *argsub = NULL;
  name_len = strlen(name);
  cmp = strncmp(argv, name, name_len);
  if (cmp != 0) {
    return cmp;
  }
  argv_len = strlen(argv);
  if (argv_len > name_len) {
    if (*(argv+name_len) == '=') {
      *argsub = (char *)argv+name_len+1;
    } else {
      cmp = 1;
    }
  }
  return cmp;
}

static char *get_param(int argc, char * const argv[], int *pindex, char *argsub)
{
  char *p = NULL;
  if (argsub != NULL) {
    return argsub;
  }
  if (++(*pindex) < argc) {
    p = argv[*pindex];
  }
  return p;
}

static int optionEntry_save_param(OptionEntry *opt, const char *param)
{
  int rc = FALSE;

  switch (opt->arg) {
  case OPTION_ARG_NONE:
    rc = optionEntry_save_int(opt, TRUE);
    break;
  case OPTION_ARG_STRING:
    rc = optionEntry_save_arg_string(opt, param);
    break;
  case OPTION_ARG_INT:
    rc = optionEntry_save_arg_int(opt, param);
    break;
  case OPTION_ARG_LONG:
    rc = optionEntry_save_arg_long(opt, param);
    break;
  case OPTION_ARG_FLOAT:
    rc = optionEntry_save_arg_float(opt, param);
    break;
  case OPTION_ARG_DOUBLE:
    rc = optionEntry_save_arg_double(opt, param);
    break;
  default:
    /* error bad type */
    rc = FALSE;
    break;
  }
  return rc;
}

static int optionEntry_save_int(OptionEntry *opt, int v)
{
  if (optionEntry_can_save(opt)) {
    *((int *)opt->save) = v;
  }
  return TRUE;
}

static int optionEntry_save_arg_string(OptionEntry *opt, const char *arg)
{
  if (optionEntry_can_save(opt)) {
    *((char **)opt->save) = (char *)arg;
  }
  return TRUE;
}

static int optionEntry_save_arg_int(OptionEntry *opt, const char *arg)
{
  char *end;
  long v;
  v = strtol(arg, &end, 10);
  if ((arg == end) || (*end != '\0')) {
    /* error bad number */
    return FALSE;
  }
  if (optionEntry_can_save(opt)) {
    *((int *)opt->save) = v;
  }
  return TRUE;
}

static int optionEntry_save_arg_long(OptionEntry *opt, const char *arg)
{
  char *end;
  long v;
  v = strtol(arg, &end, 10);
  if ((arg == end) || (*end != '\0')) {
    /* error bad number */
    return FALSE;
  }
  if (optionEntry_can_save(opt)) {
    *((long *)opt->save) = v;
  }
  return TRUE;
}

static int optionEntry_save_arg_float(OptionEntry *opt, const char *arg)
{
  char *end;
  float v;
  v = strtof(arg, &end);
  if ((arg == end) || (*end != '\0')) {
    /* error bad number */
    return FALSE;
  }
  if (optionEntry_can_save(opt)) {
    *((float *)opt->save) = v;
  }
  return TRUE;
}

static int optionEntry_save_arg_double(OptionEntry *opt, const char *arg)
{
  char *end;
  double v;
  v = strtod(arg, &end);
  if ((arg == end) || (*end != '\0')) {
    /* error bad number */
    return FALSE;
  }
  if (optionEntry_can_save(opt)) {
    *((double *)opt->save) = v;
  }
  return TRUE;
}

static int optionEntry_can_save(OptionEntry *opt)
{
  if (opt == NULL) {
    return FALSE;
  }
  if (opt->save == NULL) {
    return FALSE;
  }
  return TRUE;
}
