#include <X11/keysym.h> /* XK_q */
#include <X11/Xutil.h> /* XDestroyImage(), XPutPixel() */
#include "common.h"
#include "cache.h"
#include "dboard.h"
#include "loader.h"
#include "basetime.h"
#include "option.h"
#include "image_utils.h"
#include "cmsprofile.h"
#include "cmsimage.h"
#include "output_color.h"

#define SCREEN_RGB_IMAX 256
#define TEMPERATURE 6500
#define SCREEN_GAMMA 2.2


struct xImageWork {
  int dispose_op;
  DBoardImage *previous;
};
typedef struct xImageWork XImageWork;

struct wrapper {
  unsigned int index;
  DBoard *board;
  ImageWork *work;
  XImageWork *xWork;
};

struct player {
  ImageSet *images;
  Image *current;
  Cache *cache;
  int disable_cache;
  ImageWork *imageWork;
  XImageWork xImageWork;
  int count;
  int loop_count;
  int running, running0;
  BaseTime *t0, *t;
  const char *format;
  int argc;
  const char **argv;
  int arg_index, arg_index0;
  const char *icc_input_file;
  const char *icc_output_file;
  int icc_input_use;
  int icc_input_sRGB;
  int icc_output_sRGB;
  const char *icc_rendering_intent;
  CmsProfile *profile_input;
  CmsProfile *profile_output;
  CmsProfile *profile_sRGB;
  CmsIntent intent;
};
typedef struct player Player;

static int x_main(int argc, char **argv);

static void player_init(Player *player, int argc, const char **argv, int arg_index);
static void player_free(Player *player);
static void player_image_next(Player *player);
static void player_arg_next(Player *player);
static void player_arg_previous(Player *player);
static int player_read_file(Player *player);
static void player_reset(Player *player);
static int player_has_multiframe(Player *player);

static int moving_process(DBoard *board, int *pdone, void *data);
static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data);
static void board_setup(DBoard *board, Player *player);
static void image_setup(DBoard *board, Player *player);

static unsigned int ximage_hashcode(const void *arg);
static void *ximage_get(const void *arg);
static int ximage_release(void *data);
static int dBoardImage_free_image(DBoardImage *bimage);

static DBoardImage *ximage_setup(DBoard *board, ImageWork *work, XImageWork *xWork);
static void ximage_indexed(DBoardImage *bimage, Image *image, OutputColor *color);
/*static void ximage_indexed(DBoardImage *bimage, ImageWork *work, OutputColor *color);*/
static void ximage_mapped(DBoardImage *bimage, ImageWork *work, OutputColor *color, unsigned long (*xcolor_get_pixel)(OutputColor *color, unsigned char *pp));
static void xmage_pixel(DBoardImage *bimage, ImageWork *work, DBoard *board, unsigned long (*dBoard_get_pixel)(DBoard *board, unsigned char *pp));
static unsigned long xcolor_background_pixel(DBoard *board, ImageWork *work);
static unsigned long xcolor_get_pixel_rgb8(OutputColor *output_color, unsigned char *p);
static unsigned long xcolor_get_pixel_rgb16(OutputColor *output_color, unsigned char *p);


int main(int argc, char **argv)
{
  x_main(argc, argv);
  return EXIT_SUCCESS;
}

static int x_main(int argc, char **argv)
{
  DBoard *board;
  Player player;
  OptionEntry option[] = {
    { "animate", "a", OPTION_ARG_NONE, &(player.running0) },
    { "format", NULL, OPTION_ARG_STRING, &(player.format) },
    { "disable-cache", NULL, OPTION_ARG_NONE, &(player.disable_cache) },
    { "input-no-embedded", "in", OPTION_ARG_NONE, &(player.icc_input_use) },
    { "input-profile", "ip", OPTION_ARG_STRING, &(player.icc_input_file) },
    { "input-sRGB", "is", OPTION_ARG_NONE, &(player.icc_input_sRGB) },
    { "output-profile", "op", OPTION_ARG_STRING, &(player.icc_output_file) },
    { "output-sRGB", "os", OPTION_ARG_NONE, &(player.icc_output_sRGB) },
    { "rendering-intent", "ri", OPTION_ARG_STRING, &(player.icc_rendering_intent) },
    { NULL, NULL, OPTION_ARG_NONE, NULL }
  };

  board = dBoard_open("xiview", &argc, argv);
  if (board == NULL) {
    fprintf(stderr, "Cannot open toolkit.\n");
    exit(EXIT_FAILURE);
  }
  player.running0 = FALSE;
  player.format = NULL;
  player.disable_cache = FALSE;
  player.icc_input_use = FALSE;
  player.icc_input_file = NULL;
  player.icc_input_sRGB = FALSE;
  player.icc_output_file = NULL;
  player.icc_output_sRGB = FALSE;
  player.icc_rendering_intent = NULL;
  argc--;
  argv++;
  if (!option_get(&argc, &argv, option)) {
    /* usage */
    exit(EXIT_FAILURE);
  }
  player_init(&player, argc, argv, 0);

  if (!player_read_file(&player)) {
    fprintf(stderr, "Cannot read file.\n");
    player_free(&player);
    exit(EXIT_FAILURE);
  }
  board_setup(board, &player);
  image_setup(board, &player);
  dBoard_setHandler(board, handle_event,
                    ExposureMask |
                    KeyPressMask | KeyReleaseMask | ColormapChangeMask);
  dBoard_setProcessor(board, moving_process);
  dBoard_show(board);
  while (!dBoard_main(board, &player)) {
    ;
  }
  dBoard_close(board);
  player_free(&player);
  return TRUE;
}

static void player_init(Player *player, int argc, const char **argv, int arg_index)
{
  player->t0 = baseTime_new();
  player->t = baseTime_new();
  player->argc = argc;
  player->argv = argv;
  player->arg_index = player->arg_index0 = arg_index;
  player->images = imageSet_new();
  player->imageWork = imageWork_alloc();
  player->cache = NULL;
  player->profile_input = NULL;
  if (player->icc_input_file != NULL) {
    player->profile_input = cmsProfile_open_file(player->icc_input_file);
  }
  player->profile_output = NULL;
  if (player->icc_output_file != NULL) {
    player->profile_output = cmsProfile_open_file(player->icc_output_file);
  } else {
#if 0
    int tempK = TEMPERATURE;
    double gamma[3];
    for (int i = 0; i < 3; i++) {
      gamma[i] = 1.0/SCREEN_GAMMA;
    }
    player->profile_output = cmsProfile_open_temp(tempK, gamma, 3);
#endif /* 0 */
    player->profile_output = NULL;
  }
  player->intent = CMS_INTENT_PERCEPTUAL;
  if (player->icc_rendering_intent != NULL) {
    CmsIntent intent = CMS_INTENT_PERCEPTUAL;
    switch (*(player->icc_rendering_intent)) {
    case 'P': /* perceptual */
      intent = CMS_INTENT_PERCEPTUAL;
      break;
    case 'A': /* absolute colorimetric */
      intent = CMS_INTENT_ABSOLUTE_COLORIMETRIC;
      break;
    case 'R': /* relative colorimetric */
      intent = CMS_INTENT_RELATIVE_COLORIMETRIC;
      break;
    case 'S': /* saturation */
      intent = CMS_INTENT_SATURATION;
      break;
    default:
      break;
    }
    player->intent = intent;
  }
  player->profile_sRGB = cmsProfile_open_sRGB();
}

static void player_free(Player *player)
{
  baseTime_delete(player->t0);
  baseTime_delete(player->t);
  imageSet_free(player->images);
  XFREE(player->images);
  imageWork_free(player->imageWork);
  cache_close(player->cache);
  cmsProfile_close(player->profile_input);
  cmsProfile_close(player->profile_output);
  cmsProfile_close(player->profile_sRGB);
}

static void player_image_next(Player *player)
{
  Image *image;
  int num_loop;

  image = player->current;
  if (image_next(image) == NULL) {
    player->current = imageSet_get_first(player->images);
    player->count = 0;
    player->loop_count++;
  } else {
    player->current = image_next(image);
    player->count++;
  }
  num_loop = imageSet_get_num_plays(player->images);
  if (num_loop > 0) {
    if (player->loop_count >= num_loop) {
      player->running = FALSE;
    }
  }
}

static void player_arg_next(Player *player)
{
  if (++(player->arg_index) >= player->argc) {
    player->arg_index = player->argc-1;
  }
}

static void player_arg_previous(Player *player)
{
  if(--(player->arg_index) < 1) {
    player->arg_index = player->arg_index0;
  }
}

static int player_read_file(Player *player)
{
  Entry entry;
  const char *filename;
  const char *format;
  int index, rc;

  index = player->arg_index;
  filename = (player->argv)[index];
  format = player->format;
  DEBUG_PRINT("filename = '%s'\n", filename);
  entry.filename = filename;
  cache_close(player->cache);
  imageSet_free(player->images);
  rc = load_images(&entry, player->images, format);
  player_reset(player);
  if (!rc) {
    return FALSE;
  }
  return TRUE;
}

static void player_reset(Player *player)
{
  player->current = imageSet_get_first(player->images);
  player->count = 0;
  player->loop_count = 0;
  if (player_has_multiframe(player)) {
    player->running = player->running0;
    if (!player->disable_cache) {
      player->cache = cache_open(imageSet_get_num_frames(player->images), ximage_hashcode, ximage_get, ximage_release);
    } else {
      player->cache = NULL;
    }
  } else {
    player->running = FALSE;
    player->cache = NULL;
  }
  baseTime_get(player->t0);
}

static int player_has_multiframe(Player *player)
{
  return (imageSet_get_num_frames(player->images) > 1);
}

static int moving_process(DBoard *board, int *pdone, void *data)
{
  Player *player;
  BaseTime *t;
  int dt, delay;

  player = (Player *)data;
  if (!player->running) {
    baseTime_sleep_msec(20);
    return TRUE;
  }
  delay = image_get_delay(player->current);
  t = player->t;
  baseTime_get(t);
  dt = baseTime_subtract_msec(t, player->t0);
  if (dt >= delay) {
    player_image_next(player);
    image_setup(board, player);
    dBoard_copy(board);
    baseTime_get(t);
    dt = baseTime_subtract_msec(t, player->t0);
    baseTime_set(player->t0, t);
    DEBUG_PRINT("dt = %d, delay = %d\n", dt, delay);
  } else {
    int wait = (delay-dt)/10;
    if (wait < 20) {
      baseTime_sleep_msec(wait);
    } else {
      baseTime_sleep_msec(20);
    }
  }
  return TRUE;
}

static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data)
{
  Player *player;
  int rc, done;
  Atom delwin;
  KeySym keysym;

  player = (Player *)data;
  rc = done = FALSE;
  switch (event->type) {
  case Expose:
    DEBUG_PRINT("Expose\n");
    if ((event->xexpose).count == 0 ) {
      dBoard_copy(board);
      dBoard_flush(board);
    }
    break;
  case NoExpose:
    DEBUG_PRINT("NoExpose\n");
    break;
  case ClientMessage:
    DEBUG_PRINT("ClientMessage\n");
    delwin = dBoard_delete_window(board);
    if ((event->xclient).data.l[0] == delwin) {
      done = TRUE;
    }
    break;
  case KeyRelease:
    keysym = XLookupKeysym(&(event->xkey), 0);
    if (keysym == XK_n) {
      if (player_has_multiframe(player)) {
        player_image_next(player);
        image_setup(board, player);
        dBoard_copy(board);
        dBoard_flush(board);
      }
    } else if (keysym == XK_space) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_next(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_BackSpace) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_previous(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_s) {
      if (player_has_multiframe(player)) {
        player->running = !(player->running);
      }
    } else if (keysym == XK_q) {
      done = TRUE;
    } else {
      ;
    }
    break;
  case DestroyNotify:
    DEBUG_PRINT("DestroyNotify\n");
    done = TRUE;
    break;
  default:
    DEBUG_PRINT("event: %d\n", event->type);
    break;
  }
  *pdone = done;
  return rc;
}

static void board_setup(DBoard *board, Player *player)
{
  int width, height;
  int index;
  const char *filename;

  imageSet_get_size(player->images, &width, &height);
  dBoard_reset(board, width, height);
  index = player->arg_index;
  filename = (player->argv)[index];
  dBoard_set_name(board, filename);

  BgPrefer bg_prefer;
  unsigned short bg_r, bg_g, bg_b;
  unsigned char bgdata[6];
  int bg_setted;
  int depth, w, h;
  int use_input;
  CmsProfile *profile_in;
  CmsProfile *profile_out;

  use_input = player->icc_input_use;
  profile_in = (player->icc_input_sRGB) ? player->profile_sRGB : player->profile_input;
  profile_out = (player->icc_output_sRGB) ? player->profile_sRGB : player->profile_output;
  bg_r = bg_g = bg_b = 0x7f7f;
  dBoard_get_attributes(board, &depth, &w, &h, &bg_r, &bg_g, &bg_b, &bg_setted);
  DEBUG_PRINT(" attr: %d [%dx%d] bg_setted=%s (%x,%x,%x)\n",
              depth, w, h, (bg_setted) ? "TRUE" : "FALSE", bg_r, bg_g, bg_b);
  imageWork_init(player->imageWork, player->images);
  imageWork_set_profiles(player->imageWork, profile_in, profile_out, player->intent, use_input);
  bg_prefer = (bg_setted) ? BG_USER : BG_IMAGE;

  bgdata[0] = (bg_r >> 8) & 0xff;
  bgdata[1] = bg_r & 0xff;
  bgdata[2] = (bg_g >> 8) & 0xff;
  bgdata[3] = bg_g & 0xff;
  bgdata[4] = (bg_b >> 8) & 0xff;
  bgdata[5] = bg_b & 0xff;
#if 0
  set_rgb16(bgdata, bg_r, bg_g, bg_b);
#endif /* 0 */
  imageWork_set_background(player->imageWork, bgdata, bg_prefer);
}

#if 0
static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b)
{
  set_c16(&rgb[0], r);
  set_c16(&rgb[2], g);
  set_c16(&rgb[4], b);
}

static void set_c16(unsigned char *c16, unsigned short c)
{
  c16[0] = (c >> 8) & 0xff;
  c16[1] = c & 0xff;
}
#endif /* 0 */

static void image_setup(DBoard *board, Player *player)
{
  DBoardImage *bimage;
  DBoardImage *last_frame;
  XImageWork *xWork;
  ImageWork *imageWork;
  Image *image;
  int dispose_op;

  image = player->current;
  if (image == NULL) {
    return;
  }
  imageWork = player->imageWork;
  xWork = &(player->xImageWork);
  if (player->cache != NULL) {
    struct wrapper t;
    DBoardImage *cached;

    t.index = player->count;
    cached = (DBoardImage *)cache_tryget(player->cache, &t);
    if (cached == NULL) {
      int w, h;
      last_frame = NULL;
      dispose_op = image_get_dispose_op(image);
      DEBUG_PRINT(" dispose_op = %d\n", dispose_op);
      switch (dispose_op) {
      case DISPOSE_OP_PREVIOUS:
        imageSet_get_size(player->images, &w, &h);
        last_frame = dBoard_get_image(board, 0, 0, w, h);
        break;
      }
      imageWork_update(imageWork, image);
      t.board = board;
      t.work = imageWork;
      t.xWork = xWork;
      cached = (DBoardImage *)cache_get(player->cache, &t);
    }
    bimage = cached;
  } else {
    int w, h;
    last_frame = NULL;
    dispose_op = image_get_dispose_op(image);
    DEBUG_PRINT(" dispose_op = %d\n", dispose_op);
    switch (dispose_op) {
    case DISPOSE_OP_PREVIOUS:
      imageSet_get_size(player->images, &w, &h);
      last_frame = dBoard_get_image(board, 0, 0, w, h);
      break;
    }
    imageWork_update(imageWork, image);
    bimage = ximage_setup(board, imageWork, xWork);
  }
  xWork->dispose_op = dispose_op;
  xWork->previous = last_frame;
  dBoard_put_image(board, bimage);
  if (player->cache == NULL) {
    dBoardImage_free_image(bimage);
  }
}

static unsigned int ximage_hashcode(const void *arg)
{
  struct wrapper *p;

  p = (struct wrapper *)arg;
  return p->index;
}

static void *ximage_get(const void *arg)
{
  struct wrapper *p;
  struct DBoardImage *cached;

  p = (struct wrapper *)arg;
  cached = ximage_setup(p->board, p->work, p->xWork);
  return cached;
}

static int ximage_release(void *p)
{
  DBoardImage *cached;
  int rc;

  cached = (DBoardImage *)p;
  rc = dBoardImage_free_image(cached);
  XFREE(cached);
  return rc;
}

static int dBoardImage_free_image(DBoardImage *bimage)
{
  int rc = TRUE;
  XImage *ximage = bimage->ximage;

  if (ximage == NULL) {
    return TRUE;
  }
  rc = XDestroyImage(ximage);
  return rc;
}

static DBoardImage *ximage_setup(DBoard *board, ImageWork *work, XImageWork *xWork)
{
  DBoardImage *bimage;
  int dispose_op, blend_op;
  int width, height;
  int xoff, yoff;
  ImageType type;
  Image *image;
  unsigned long background;
  DBoardImage *prev_bimage;


  image = imageWork_get_image(work);
  type = image_get_type(image);
  imageWork_get_updated_size(work, &width, &height, &xoff, &yoff);
  /* dispose image */
  dispose_op = imageWork_get_dispose_op(work);
  DEBUG_PRINT(" dispose_op type: %d\n", dispose_op);
  switch (dispose_op) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    background = xcolor_background_pixel(board, work);
    dBoard_fill_rectangle(board, xoff, yoff, width, height, background);
    break;
  case DISPOSE_OP_PREVIOUS:
    dBoard_put_image(board, xWork->previous);
    dBoardImage_free_image(xWork->previous);
    break;
  default:
    DEBUG_PRINT(" unknown dipose_op type %d. why?\n", dispose_op);
    break;
  }
  blend_op = image_get_blend_op(image);
  DEBUG_PRINT(" blend_op type: %d\n", blend_op);
  switch (blend_op) {
  case BLEND_OP_NONE:
    bimage = dBoard_get_image(board, xoff, yoff, width, height);
    break;
  case BLEND_OP_SOURCE:
    background = xcolor_background_pixel(board, work);
    dBoard_fill_rectangle(board, xoff, yoff, width, height, background);
    bimage = dBoard_get_image(board, xoff, yoff, width, height);
    break;
  case BLEND_OP_OVER:
    bimage = dBoard_get_image(board, xoff, yoff, width, height);
    break;
  default:
    DEBUG_PRINT(" unknown blend_op type %d. why?\n", blend_op);
    break;
  }
  DEBUG_PRINT(" image type: %d %d (%d, %d) [%dx%d]\n",
              type, image_get_num_indexes(image),
              xoff, yoff, width, height);
  DEBUG_PRINT("  (%d, %d) [%dx%d]\n",
              bimage->xoff, bimage->yoff, bimage->width, bimage->height);
  switch (type) {
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    DEBUG_PRINT("image_type_indexed. type=%d\n", type);
    image = imageWork_get_image(work);
    ximage_indexed(bimage, image, dBoard_get_colors(board, 8));
    /*ximage_indexed(bimage, work, dBoard_get_colors(board, 8));*/
    break;
  case IMAGE_TYPE_RGB_8:
    DEBUG_PRINT("image_type_rgb8.\n");
    ximage_mapped(bimage, work, dBoard_get_colors(board, 8), xcolor_get_pixel_rgb8);
    break;
  case IMAGE_TYPE_RGB_16:
    DEBUG_PRINT("image_type_rgb16.\n");
    /*ximage_pixel(bimage, work, board, dBoard_get_pixel_rgb16);*/
    ximage_mapped(bimage, work, dBoard_get_colors(board, 16), xcolor_get_pixel_rgb16);
    break;
  default:
    fprintf(stderr, "unknown image type %d. why?\n", type);
    break;
  }
  return bimage;
}

static void ximage_indexed(DBoardImage *bimage, Image *image, OutputColor *color)
/*static void ximage_indexed(DBoardImage *bimage, ImageWork *work, OutputColor *color)*/
{
  unsigned long xcolors[SCREEN_RGB_IMAX];
  unsigned char alpha[SCREEN_RGB_IMAX];
  ImageType type;
  /*Image *image;*/
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;
  int num;

  /*image = imageWork_get_image(work);*/
  type = image_get_type(image);
  bpp = imageType_bytes_per_pixel(type);
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  xoff = bimage->xoff;
  yoff = bimage->yoff;
  w = bimage->width;
  h = bimage->height;
  /*imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);*/
  num = image_get_num_indexes(image);
  y = yoff;
  switch (type) {
  case IMAGE_TYPE_INDEXED_RGB_8:
    for (int i = 0; i < num; i++) {
      unsigned char data[3];
      image_get_index(image, i, data);
      xcolors[i] = outputColor_get_pixel(color, data[0], data[1], data[2]);
    }
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp;
      for (int i = 0; i < w; i++) {
        int c;
        c = *(pp++);
        XPutPixel(bimage->ximage, i, j, xcolors[c]);
      }
      y++;
    }
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    for (int i = 0; i < num; i++) {
      unsigned char data[4];
      image_get_index(image, i, data);
      xcolors[i] = outputColor_get_pixel(color, data[0], data[1], data[2]);
      alpha[i] = data[3];
    }
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp;
      for (int i = 0; i < w; i++) {
        int c;
        c = *(pp++);
        if (alpha[c] >= 0x80) {
          /* fully opaque */
          XPutPixel(bimage->ximage, i, j, xcolors[c]);
        }
      }
      y++;
    }
    break;
  }
}

static void ximage_mapped(DBoardImage *bimage, ImageWork *work, OutputColor *color, unsigned long (*xcolor_get_pixel)(OutputColor *color, unsigned char *pp))
{
  Image *image;
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  image = imageWork_get_image(work);
  bpp = imageType_bytes_per_pixel(image_get_type(image));
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);
  y = yoff;
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*y+xoff)*bpp;
    for (int i = 0; i < w; i++) {
      unsigned long xcol;
      xcol = (*xcolor_get_pixel)(color, pp);
      XPutPixel(bimage->ximage, i, j, xcol);
      pp += bpp;
    }
    y++;
  }
}

static void ximage_pixel(DBoardImage *bimage, ImageWork *work, DBoard *board, unsigned long (*dBoard_get_pixel)(DBoard *board, unsigned char *pp))
{
  Image *image;
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  image = imageWork_get_image(work);
  bpp = imageType_bytes_per_pixel(image_get_type(image));
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);
  y = yoff;
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*y+xoff)*bpp;
    for (int i = 0; i < w; i++) {
      unsigned long xcol;
      xcol = (*dBoard_get_pixel)(board, pp);
      XPutPixel(bimage->ximage, i, j, xcol);
      pp += bpp;
    }
    y++;
  }
}

#if 0
imageWork_get_updated_image(work, &pp1, &bpp, &w0, &h0, &w, &h, &xoff, &yoff);

static void imageWork_get_updated_image(ImageWork *work, unsigned char **pdata, int *pbpp, int *pw0, int *ph0, int *pw, int *ph, int *pxoff, int *pyoff)
{
  Image *image;

  image = imageWork_get_image(work);
  *pbpp = imageType_bytes_per_pixel(image_get_type(image));
  *pdata = image_get_data(image);
  image_get_size(image, pw0, ph0);
  imageWork_get_updated_size(work, pw, ph, pxoff, pyoff);
}
#endif /* 0 */

static unsigned long xcolor_background_pixel(DBoard *board, ImageWork *work)
{
  const unsigned char unknown[] = { 0x7f, 0x7f, 0x7f };
  unsigned long pixel;
  unsigned char data[8];
  ImageType type;

  imageWork_get_background(work, data, &type);
  switch (type) {
  case IMAGE_TYPE_RGB_8:
    pixel = xcolor_get_pixel_rgb8(dBoard_get_colors(board, 8), data);
    break;
  case IMAGE_TYPE_RGB_16:
    /*pixel = dBoard_get_pixel_rgb16(board, data);*/
    pixel = xcolor_get_pixel_rgb16(dBoard_get_colors(board, 16), data);
    break;
  case IMAGE_TYPE_UNKNOWN:
    pixel = xcolor_get_pixel_rgb8(dBoard_get_colors(board, 8), unknown);
    DEBUG_PRINT(" background type = unknown\n");
    break;
  default:
    DEBUG_PRINT(" background type = %d, why?\n", type);
    break;
  }
  DEBUG_PRINT(" pixel = %x\n", pixel);
  return pixel;
}

static unsigned long xcolor_get_pixel_rgb8(OutputColor *output_color, unsigned char *p)
{
  unsigned char r, g, b;
  r = *p++;
  g = *p++;
  b = *p++;
  return outputColor_get_pixel(output_color, r, g, b);
}

static unsigned long xcolor_get_pixel_rgb16(OutputColor *output_color, unsigned char *p)
{
  unsigned short c16[3];
  int shift;

  for (int i = 0; i < 3; i++) {
    c16[i]  = (*(p++) << 8) & 0xff00;
    c16[i] |= *(p++) & 0xff;
  }
  return outputColor_get_pixel(output_color, c16[0], c16[1], c16[2]);
}

#if 0
/*
 * bit_depth  Number of significant figures
 */
static unsigned long xcolor_get_pixel_rgb(OutputColor *output_color, unsigned char *p, int bit_depth)
{
  unsigned long pixel;

  if (bit_depth <= 8) {
    ;
  } else if ((bit_depth > 8) && (bit_depth <= 16)) {
    unsigned short c[3];
    unsigned short mask;

    mask = (0xffff >> (16-bit_depth));
    for (int i = 0; i < 3; i++) {
      unsigned short c16;
      c16  = (*(p++) << 8) & 0xff00;
      c16 |= *(p++) & 0xff;
      c[i] = (c16 >> (16-bit_depth)) & mask;
    }
    pixel = outputColor_get_pixel(output_color, c[0], c[1], c[2]);
  } else if (bit_depth > 16) {
    fprintf(stderr, "xcolor_get_pixel_rgb() is not supported bit_depth %d.\n", bit_depth);
  }
  return pixel;
}
#endif /* 0 */
